/*
 * PowerForms
 * Copyright (C) 1999-2002 BRICS.
 *
 * Author: Mikkel Ricky <ricky@brics.dk>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the  Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 * USA.
 *
 * Home page: http://www.brics.dk/bigwig/powerforms/
 */

package dk.brics.powerforms;

import java.io.*;
import java.net.*;
import java.util.*;

abstract class Constraint extends PwfElement {
  static final int FIELD = 0;
  static final int GLOBAL = 1;
  static final int TEST = 2;
  protected HtmlStatus status = null;

  HtmlStatus getStatus() {
    return status;
  }

  void setStatus(HtmlStatus status) throws TransformException {
    if (this.status != null) {
      status.reportError("constraint `"+id+"' already has a status element");
    } else {
      this.status = status;
    }
  }

  void scriptDefine(HtmlDocument document, PrintWriter out) {}
  void scriptInit(HtmlDocument document, PrintWriter out) {}
  void code(HtmlDocument document, PrintWriter out) {}

  abstract boolean validate(Map values, ErrorHandler handler);
}

class ConstraintApply extends Constraint {
  private static int number = 0;

  private Constraint constraint;
  private AutoComplete autocomplete;
  /* private StringPairSet constrains; */
  private StringPair constrains;
  String formName, fieldName;
  int index;

  ConstraintApply(Attributes attributes, Constraint constraint) {
    this.constraint = constraint;
    index = number++;
    PwfContent content = new PwfContent();
    content.add(constraint);
    init("constraint", attributes, content);
  }

  void addAutoComplete(AutoComplete autocomplete) throws TransformException {
    if (this.autocomplete != null) {
      reportError("multiple auto completions for constraint `"+id+"'");
    }
    this.autocomplete = autocomplete;
  }

  AutoComplete getAutoComplete() {
    return autocomplete;
  }

  void addDependency(String form, String field) {
    getPowerForms().addDependency(form, field, this);
  }

  StringPair getConstrains() {
    return constrains;
  }

  void check() throws TransformException {
    formName = attributes.getValue("form");
    fieldName = attributes.getValue("field");
    constrains = new StringPair(formName, fieldName);
    getPowerForms().addConstraintApply(this);
    super.check();
  }

  String getScriptName() {
    return "pwf_constraint_"+index;
  }

  void scriptDefine(HtmlDocument document, PrintWriter out) {
    out.println();
    out.println("// ("+formName+", "+fieldName+")");
    out.println("function "+getScriptName()+"() {");
    /* out.println("PowerForms.DEBUG('"+getScriptName()+"');"); */
    out.println("  var accept = true;");
    out.println("  var change = false;");
    out.println("  var alert = null;");
    out.println("  var error = null;");
    out.println("  var messages = new Array();");
    out.println("  var state, element;");
    /* out.println("  var constraint = PowerForms.IGNORE;"); */
    constraint.script(document, out);
    out.println();
    if (status != null) {
      out.println(status.getScriptName()+".update(state);");
    }
    out.println("  return Array(accept, change, messages);");
    out.println("}");
  }

  boolean validate(Map values, ErrorHandler handler) {
    return constraint.validate(values, handler);
  }

  public String toString() {
    return "["+tagName+" apply "+index+"; constrains: "+constrains+"]";
  }
}

class ConstraintDef extends Constraint {
  private Constraint constraint;

  ConstraintDef(Attributes attributes, Constraint constraint) {
    this.constraint = constraint;
    PwfContent content = new PwfContent();
    content.add(constraint);
    init("constraintdef", attributes, content);
  }

  boolean validate(Map values, ErrorHandler handler) {
    return constraint.validate(values, handler);
  }
}

class ConstraintRef extends Constraint {
  private String idref;
  private Constraint constraint;

  ConstraintRef(Attributes attributes) {
    init("constraint", attributes);
  }

  void check() throws TransformException {
    idref = attributes.requireValue("idref");
    if (idref != null) {
      constraint = getPowerForms().getConstraint(idref);
      if (constraint == null) {
	reportError("constraint `"+idref+"' not definined");
      }
    }
  }

  boolean validate(Map values, ErrorHandler handler) {
    return constraint.validate(values, handler);
  }
}

class ConstraintIf extends Constraint {
  private Expression test;
  private Constraint thenpart, elsepart;

  ConstraintIf(Expression test, Constraint thenpart, Constraint elsepart) {
    this.test = test;
    this.thenpart = thenpart;
    this.elsepart = elsepart;
    PwfContent content = new PwfContent();
    content.add(test);
    content.add(thenpart);
    content.add(elsepart);
    init("if", content);
  }

  void check() throws TransformException {
    test.check(true);
    thenpart.check();
    elsepart.check();
  }

  void script(HtmlDocument document, PrintWriter out) {
    out.print("if (PowerForms.toBoolean(");
    test.script(document ,out);
    out.print(")) {");
    out.println();
    thenpart.script(document, out);
    out.print("} else {");
    out.println();
    elsepart.script(document, out);
    out.println("}");
  }

  boolean validate(Map values, ErrorHandler handler) {
    if (test.evaluate(values)) {
      return thenpart.validate(values, handler);
    } else {
      return elsepart.validate(values, handler);
    }
  }
}

class ConstraintIgnore extends Constraint {
  private ConstraintApply constraint;
  private String formName;
  private String fieldName;

  ConstraintIgnore(Attributes attributes) {
    init("ignore", attributes);
  }

  void check() throws TransformException {
    constraint = (ConstraintApply)getAncestor(ConstraintApply.class);
    formName = constraint.formName;
    if (constraint.fieldName != null) {
      fieldName = constraint.fieldName;
      attributes.disallowAttribute("field");
    } else {
      fieldName = attributes.requireValue("field");
    }
  }

  void script(HtmlDocument document, PrintWriter out) {
    out.println("// Ignore ...");
    HtmlForm form = document.getForm(formName);
    HtmlFieldList fieldList = (form == null ? null : form.getFieldList(fieldName));
    if (fieldList != null) {
      for (int i = 0; i < fieldList.size(); i++) {
	HtmlField field = fieldList.getField(i);
	if ((constraint.fieldName != null) && (field.getStatus() != null)) {
	  out.println(field.getStatus().getScriptName()+".update(PowerForms.NA);");
	}
      }
    }
  }

  boolean validate(Map values, ErrorHandler handler) {
    return true;
  }
}

class ConstraintExpression extends Constraint {
  private Expression expression;

  ConstraintExpression(Expression expression) {
    this.expression = expression;
    PwfContent content = new PwfContent();
    content.add(expression);
    init("constraint-expression", content);
  }

  boolean validate(Map values, ErrorHandler handler) {
    return expression.validate(values, handler);
  }
}
