/*
 * PowerForms
 * Copyright (C) 1999-2002 BRICS.
 *
 * Author: Mikkel Ricky <ricky@brics.dk>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the  Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 * USA.
 *
 * Home page: http://www.brics.dk/bigwig/powerforms/
 */

package dk.brics.powerforms;

import java.io.*;
import java.net.*;
import java.util.*;

abstract class Expression extends PwfElement {
  boolean isTest = false;
  ConstraintApply constraint;

  void setConstraint() {
    constraint = (ConstraintApply)getAncestor(ConstraintApply.class);
  }

  void check() throws TransformException {
    check(false);
  }

  void check(boolean isTest) throws TransformException {
    this.isTest = isTest;
    setConstraint();
    if (content != null) {
      for (int i = 0; i < content.size(); i++) {
	PwfElement e = content.get(i);
	if (e instanceof Expression) {
	  ((Expression)e).check(isTest);
	} else {
	   e.check();
	}
      }
    }
  }

  void scriptDefine(HtmlDocument document, PrintWriter out) {}

  void scriptInit(HtmlDocument document, PrintWriter out) {}

  void code(PrintWriter out) {}

  private String getMessage(String name) {
    String message = (attributes == null ? null : attributes.getValue(name));
    if (message == null) {
      for (PwfElement p = parent; p != null; p = p.parent) {
	message = (p.attributes == null ? null : p.attributes.getValue(name));
	if (message != null) {
	  return message;
	}
      }
    }
    return message;
  }

  protected String getErrorMessage(String defaultValue) {
    String message = getMessage("error");
    return (message != null ? message : defaultValue);
  }

  protected String getWarningMessage(String defaultValue) {
    String message = getMessage("warning");
    return message != null ? message : getErrorMessage(defaultValue);
  }

  protected String getHelpMessage(String defaultValue) {
    String message = getMessage("help");
    return (message != null ? message : defaultValue);
  }

  abstract boolean evaluate(Map values);
  abstract boolean validate(Map values, ErrorHandler handler);
}

class ExpressionCount extends Expression {
  String formName, fieldName;
  private int min, max;

  ExpressionCount(Attributes attributes) {
    init("count", attributes);
  }

  void check(boolean isTest) throws TransformException {
    super.check(isTest);
    if (attributes.getValue("number") != null) {
      min = max = attributes.getIntValue("number", 0);
    } else {
      min = attributes.getIntValue("min", 0);
      max = attributes.getIntValue("max", -1);
    }
    formName = constraint.formName;
    if (isTest || (constraint.fieldName == null)) {
      fieldName = attributes.requireValue("field");
      constraint.addDependency(formName, fieldName);
    } else {
      fieldName = constraint.fieldName;
      attributes.disallowAttribute("field");
    }
  }

  void script(HtmlDocument document, PrintWriter out) {
    HtmlForm form = document.getForm(formName);
    HtmlFieldList fieldList = (form == null ? null : form.getFieldList(fieldName));
    if (isTest) {
      if (fieldList == null) {
	out.print("PowerForms.NO");
      } else {
	out.print("PowerForms.count("+form.index+", '"+fieldList.name+"', "+min+", "+(max < 0 ? "Infinity" : String.valueOf(max))+")");
      }
    } else {
      if (fieldList != null) {
	out.println("if (!PowerForms.toBoolean(PowerForms.count("+form.index+", '"+fieldList.name+"', "+min+", "+(max < 0 ? "Infinity" : String.valueOf(max))+"))) {");
	/*
	  out.println("  if (pwf_event != null) {");
	  out.println("    if (pwf_event.preventDefault) {");
	  out.println("      pwf_event.preventDefault();");
	  out.println("    } else {");
	  out.println("      pwf_event.cancel();");
	  out.println("    }");
	  out.println("  }");
	*/
	out.println("  PowerForms.addMessage(messages, (PowerForms.getValues("+form.index+", '"+fieldList.name+"').size() < "+min+" ? "
		    +Html.scriptQuote(getErrorMessage(getPowerForms().validationError("powerforms.tooFewValues", fieldName)))+" : "
		    +Html.scriptQuote(getErrorMessage(getPowerForms().validationError("powerforms.tooManyValues", fieldName)))+"));");
	out.println("  accept = false;");
	out.println("}");
      }
    }
  }

  boolean evaluate(Map values) {
    return validate(values, null);
  }

  boolean validate(Map values, ErrorHandler handler) {
    Collection c = HtmlForm.getValues(values, fieldName);
    int size = (c == null ? 0 : c.size());
    if (size < min) {
      if (handler != null) {
	handler.error(getErrorMessage(getPowerForms().validationError("powerforms.tooFewValues", fieldName)));
      }
      return false;
    } else if (max < size) {
      if (handler != null) {
	handler.error(getErrorMessage(getPowerForms().validationError("powerforms.tooManyValues", fieldName)));
      }
      return false;
    }
    return true;
  }
}

class ExpressionMatch extends Expression {
  private RegExp regexp;
  String formName, fieldName;

  ExpressionMatch(Attributes attributes, RegExp regexp) {
    this.regexp = regexp;
    PwfContent content = new PwfContent();
    content.add(regexp);
    init("match", attributes, content);
  }

  void check(boolean isTest) throws TransformException {
    super.check(isTest);
    getPowerForms().addRegExp(regexp);
    formName = constraint.formName;
    if (isTest || (constraint.fieldName == null)) {
      fieldName = attributes.requireValue("field");
      constraint.addDependency(formName, fieldName);
    } else {
      fieldName = constraint.fieldName;
      attributes.disallowAttribute("field");
    }
  }

  void script(HtmlDocument document, PrintWriter out) {
    HtmlForm form = document.getForm(formName);
    HtmlFieldList fieldList = (form == null ? null : form.getFieldList(fieldName));
    if (isTest) {
      if (fieldList == null) {
	out.print("PowerForms.YES");
      } else {
	out.print("PowerForms.match("+form.index+", '"+fieldList.name+"', "+regexp.getDfaName()+")");
      }
    } else {
      if (fieldList != null) {
	/* out.println("state = PowerForms.match("+form.index+", '"+fields[0].name+"', "+regexp.getDfaName()+");"); */
	for (int i = 0; i < fieldList.size(); i++) {
	  HtmlField field = fieldList.getField(i);
	  out.println("element = document.forms["+form.index+"].elements["+field.index+"];");
	  if (field instanceof HtmlInput) {
	    HtmlInput input = (HtmlInput)field;
	    if ((input.type == HtmlInput.TEXT) || (input.type == HtmlInput.PASSWORD)) {
	      String helpMessage = getHelpMessage("");
	      String errorMessage = getErrorMessage("Invalid input in field "+field.name);
	      String warningMessage = getWarningMessage("Incomplete input in field "+field.name);
	      out.println("state = "+regexp.getDfaName()+".run(element.value);");
	      out.println("switch (state) {");
	      out.println("case PowerForms.CRASH:");
	      out.println("  PowerForms.addMessage(messages, "+Html.scriptQuote(errorMessage)+");");
	      out.println("  accept = false;");
	      out.println("  break;");
	      out.println("ca"+"se PowerForms.REJECT:");
	      out.println("  PowerForms.addMessage(messages, "+Html.scriptQuote(warningMessage)+");");
	      out.println("  accept = false;");
	      out.println("  break;");
	      /*
		out.println("default:");
		out.println("  window.status = "+scriptQuote(getAttribute(this, "help", ""))+";");
		out.println("  break;");
	      */
	      out.println("}");
	      if (constraint.fieldName != null) {
		HtmlStatus status = field.getStatus();
		if (status != null) {
		  out.println(status.getScriptName()+".update(state, "+Html.scriptQuote(helpMessage)+", "+Html.scriptQuote(errorMessage)+", "+Html.scriptQuote(warningMessage)+");");
		}
		AutoComplete autocomplete = field.getAutoComplete();
		if (autocomplete != null) {
		  out.println("if (state == PowerForms.REJECT) {");
		  out.print(autocomplete.getScriptName()+".setCompletion("+regexp.getDfaName()+".complete");
		  switch (autocomplete.type) {
		  case AutoComplete.FIRST:
		    out.print("First(element.value)");
		    break;
		  case AutoComplete.PREFIX:
		  case AutoComplete.AGGRESSIVE:
		    out.print("Unique(element.value, true)");
		    break;
		  default:
		    out.print("Unique(element.value, false)");
		    break;
		  }
		  out.println(");");
		  if (autocomplete.type == AutoComplete.AGGRESSIVE) {
		    out.println(autocomplete.getScriptName()+".complete();");
		  }
		  out.println("} else {");
		  out.println(autocomplete.getScriptName()+".setCompletion(null);");
		  out.println("}");
		}
	      }
	    } else if ((input.type == HtmlInput.RADIO) || (input.type == HtmlInput.CHECKBOX)) {
	      if (constraint.fieldName != null) {
		out.println("change |= element.filter("+regexp.getDfaName()+".filter);");
	      } else {
		String errorMessage = getErrorMessage("Invalid input in field "+field.name);
		out.println("if ((element.checked) && !PowerForms.toBoolean("+regexp.getDfaName()+".run(element.value))) {");
		out.println("  PowerForms.addMessage(messages, "+Html.scriptQuote(errorMessage)+");");
		out.println("  accept = false;");
		out.println("}");
	      }
	    }
	  } else if (field instanceof HtmlSelect) {
	    if (constraint.fieldName != null) {
	      out.println("change |= element.filter("+regexp.getDfaName()+".filter);");
	    } else {
	      String errorMessage = getErrorMessage("Invalid input in field "+field.name);
	      out.println("if (!element.acceptValues("+regexp.getDfaName()+")) {");
	      out.println("  PowerForms.addMessage(messages, "+Html.scriptQuote(errorMessage)+");");
	      out.println("  accept = false;");
	      out.println("}");
	    }
	  }
	}
      }
    }
  }

  boolean evaluate(Map values) {
    return validate(values, null);
  }

  boolean validate(Map values, ErrorHandler handler) {
    Collection c = HtmlForm.getValues(values, fieldName);
    boolean error = false;
    if (c != null) {
      for (Iterator i = c.iterator(); i.hasNext(); ) {
	String value = i.next().toString();
	switch (regexp.run(value)) {
	case RegExp.CRASH:
	  if (handler != null) {
	    handler.error(getErrorMessage(getPowerForms().validationError("powerforms.invalidValue", value, fieldName)));
	  }
	  error = true;
	  break;
	case RegExp.REJECT:
	  if (handler != null) {
	    handler.error(getWarningMessage(getPowerForms().validationError("powerforms.incompleteValue", value, fieldName)));
	  }
	  error = true;
	  break;
	default:
	  return true;
	}
      }
    }
    return !error;
  }
}

class ExpressionEqual extends Expression {
  String value;
  String formName;
  String[] fieldNames;

  ExpressionEqual(Attributes attributes, PwfContent content) {
    init("equal", attributes, content);
  }

  void check(boolean isTest) throws TransformException {
    super.check(isTest);
    formName = constraint.formName;
    if (isTest || (constraint.fieldName == null)) {
      String fieldName = attributes.getValue("field");
      if (fieldName != null) {
	fieldNames = new String[] { fieldName };
      } else {
	fieldNames = new String[content.size()];
	if (fieldNames.length == 0) {
	  reportError("missing fields for `"+tagName+"'");
	} else {
	  for (int i = 0; i < content.size(); i++) {
	    ExpressionField field = (ExpressionField)content.get(i);
	    fieldNames[i] = field.name;
	  }
	}
      }
      for (int i = 0; i < fieldNames.length; i++) {
	constraint.addDependency(formName, fieldNames[i]);
      }
    } else {
      String fieldName = attributes.getValue("field");
      if (fieldName != null) {
	fieldNames = new String[] { constraint.fieldName, fieldName };
	constraint.addDependency(formName, fieldName);
      } else {
	fieldNames = new String[1+content.size()];
	fieldNames[0] = constraint.fieldName;
	for (int i = 0; i < content.size(); i++) {
	  ExpressionField field = (ExpressionField)content.get(i);
	  fieldNames[i+1] = field.name;
	}
      }
    }
    if (fieldNames.length > 2) {
      reportError("only two fields allowed for `"+tagName+"'");
    } else if (fieldNames.length == 1) {
      value = attributes.requireValue("value");
    } else {
      attributes.disallowAttribute("value");
    }
  }

  void script(HtmlDocument document, PrintWriter out) {
    HtmlForm form = document.getForm(formName);
    if (isTest) {
      if (fieldNames.length == 1) {
	HtmlFieldList fieldList = (form == null ? null : form.getFieldList(fieldNames[0]));
	if (fieldList == null) {
	  out.print("PowerForms.YES");
	} else {
	  out.print("PowerForms.equal("+form.index+", '"+fieldList.name+"', '"+value+"')");
	}
      } else {
	HtmlFieldList fieldList0 = form.getFieldList(fieldNames[0]);
	HtmlFieldList fieldList1 = form.getFieldList(fieldNames[1]);
	if ((fieldList0 == null) || (fieldList1 == null)) {
	  out.print("PowerForms.YES");
	} else {
	  out.print("PowerForms.equal("+form.index+", '"+fieldList0.name+"', "+form.index+", '"+fieldList1.name+"')");
	}
      }
    } else {
      if (fieldNames.length == 1) {
	HtmlFieldList fieldList = (form == null ? null : form.getFieldList(fieldNames[0]));
	if (fieldList != null) {
	  for (int i = 0; i < fieldList.size(); i++) {
	    HtmlField field = fieldList.getField(i);
	    out.println("element = document.forms["+form.index+"].elements["+field.index+"];");
	    if (field instanceof HtmlInput) {
	      HtmlInput input = (HtmlInput)field;
	      if ((input.type == HtmlInput.TEXT) || (input.type == HtmlInput.PASSWORD)) {
		String helpMessage = getHelpMessage("");
		String errorMessage = getErrorMessage("field `"+fieldList.name+"' must equal `"+value+"'");
		String warningMessage = getWarningMessage("field `"+fieldList.name+"' must equal `"+value+"'");

		out.println("state = PowerForms.equalValue(element.value, "+Html.scriptQuote(value)+");");
		out.println("switch (state) {");
		out.println("case PowerForms.NO:");
		out.println("  PowerForms.addMessage(messages, "+Html.scriptQuote(errorMessage)+");");
		out.println("  accept = false;");
		out.println("  break;");
		out.println("ca"+"se PowerForms.MAYBE:");
		out.println("  PowerForms.addMessage(messages, "+Html.scriptQuote(warningMessage)+");");
		out.println("  accept = false;");
		out.println("  break;");
		out.println("}");
		if (constraint.fieldName != null) {
		  HtmlStatus status = field.getStatus();
		  if (status != null) {
		    out.println(status.getScriptName()+".update(state);");
		  }
		  AutoComplete autocomplete = field.getAutoComplete();
		  if (autocomplete != null) {
		    out.println("if (state == PowerForms.MAYBE) {");
		    out.print(autocomplete.getScriptName()+".setCompletion(PowerForms.equalComplete");
		    switch (autocomplete.type) {
		    case AutoComplete.FIRST:
		      out.print("First(element.value, "+Html.scriptQuote(value)+")");
		      break;
		    case AutoComplete.PREFIX:
		    case AutoComplete.AGGRESSIVE:
		      out.print("Unique(element.value, "+Html.scriptQuote(value)+", true)");
		      break;
		    default:
		      out.print("Unique(element.value, "+Html.scriptQuote(value)+", false)");
		      break;
		    }
		    out.println(");");
		    if (autocomplete.type == AutoComplete.AGGRESSIVE) {
		      out.println(autocomplete.getScriptName()+".complete();");
		    }
		    out.println("} else {");
		    out.println(autocomplete.getScriptName()+".setCompletion(null);");
		    out.println("}");
		  }
		}
	      } else if ((input.type == HtmlInput.RADIO) || (input.type == HtmlInput.CHECKBOX)) {
		if (constraint.fieldName != null) {
		  out.println("if (element.checked) {");
		  out.println("  element.checked = PowerForms.toBoolean(PowerForms.equalValue(element.value, "+Html.scriptQuote(value)+"));");
		  out.println("  if (!element.checked) {");
		  out.println("    change = true;");
		  out.println("  }");
		  out.println("}");
		} else {
		  String errorMessage = getErrorMessage("Invalid input in field "+field.name);
		  out.println("if ((element.checked) && !PowerForms.toBoolean(PowerForms.equalValue(element.value, "+Html.scriptQuote(value)+"))) {");
		  out.println("  PowerForms.addMessage(messages, "+Html.scriptQuote(errorMessage)+");");
		  out.println("  accept = false;");
		  out.println("}");
		}
	      }
	    } else if (field instanceof HtmlSelect) {
	      if (constraint.fieldName != null) {
		/* out.println("change |= element.filter("+regexp.getDfaName()+".filter);"); */
	      } else {
		String errorMessage = getErrorMessage("Invalid input in field "+field.name);
		out.println("if (!element.equalValues("+Html.scriptQuote(value)+")) {");
		out.println("  PowerForms.addMessage(messages, "+Html.scriptQuote(errorMessage)+");");
		out.println("  accept = false;");
		out.println("}");
	      }
	    }
	  }
	  if (!true) {
	    out.println("if (!PowerForms.toBoolean(PowerForms.equal("+form.index+", '"+fieldList.name+"', '"+value+"'))) {");
	    out.println("  PowerForms.addMessage(messages, "+Html.scriptQuote(getErrorMessage("field `"+fieldList.name+"' must equal `"+value+"'"))+");");
	    out.println("}");
	  }
	}
      } else {
	HtmlFieldList fieldList0 = form.getFieldList(fieldNames[0]);
	HtmlFieldList fieldList1 = form.getFieldList(fieldNames[1]);
	int count = 0;
	String errorMessage = ((fieldList0 != null) && (fieldList1 != null) ? getErrorMessage("fields `"+fieldList0.name+"' and `"+fieldList1.name+"' must be equal") : null);
	String warningMessage = ((fieldList0 != null) && (fieldList1 != null) ? getWarningMessage("fields `"+fieldList0.name+"' and `"+fieldList1.name+"' must be equal") : null);
	while (count++ < 2) {
	  if ((fieldList0 != null) && (fieldList1 != null)) {
	    for (int i = 0; i < fieldList0.size(); i++) {
	      HtmlField field = fieldList0.getField(i);
	      out.println("element = document.forms["+form.index+"].elements["+field.index+"];");
	      if (field instanceof HtmlInput) {
		HtmlInput input = (HtmlInput)field;
		if ((input.type == HtmlInput.TEXT) || (input.type == HtmlInput.PASSWORD)) {
		  out.println("state = PowerForms.equalValue(element.value, "+form.index+", "+Html.scriptQuote(fieldList1.name)+");");
		  out.println("switch (state) {");
		  out.println("case PowerForms.NO:");
		  out.println("  PowerForms.addMessage(messages, "+Html.scriptQuote(errorMessage)+");");
		  out.println("  accept = false;");
		  out.println("  break;");
		  out.println("ca"+"se PowerForms.MAYBE:");
		  out.println("  PowerForms.addMessage(messages, "+Html.scriptQuote(warningMessage)+");");
		  out.println("  accept = false;");
		  out.println("  break;");
		  out.println("}");
		  if ((constraint.fieldName != null) && constraint.fieldName.equals(fieldList0.name)) {
		    HtmlStatus status = field.getStatus();
		    if (status != null) {
		      out.println(status.getScriptName()+".update(state);");
		    }
		    AutoComplete autocomplete = field.getAutoComplete();
		    if (autocomplete != null) {
		      out.println("if (state == PowerForms.MAYBE) {");
		      out.print(autocomplete.getScriptName()+".setCompletion(PowerForms.equalComplete");
		      switch (autocomplete.type) {
		      case AutoComplete.FIRST:
			out.print("First(element.value, "+form.index+", "+Html.scriptQuote(fieldList1.name)+")");
			break;
		      case AutoComplete.PREFIX:
		      case AutoComplete.AGGRESSIVE:
			out.print("Unique(element.value, "+form.index+", "+Html.scriptQuote(fieldList1.name)+", true)");
			break;
		      default:
			out.print("Unique(element.value, "+form.index+", "+Html.scriptQuote(fieldList1.name)+", false)");
			break;
		      }
		      out.println(");");
		      if (autocomplete.type == AutoComplete.AGGRESSIVE) {
			out.println(autocomplete.getScriptName()+".complete();");
		      }
		      out.println("} else {");
		      out.println(autocomplete.getScriptName()+".setCompletion(null);");
		      out.println("}");
		    }
		  }
		} else if ((input.type == HtmlInput.RADIO) || (input.type == HtmlInput.CHECKBOX)) {
		  if ((constraint.fieldName != null) && constraint.fieldName.equals(fieldList0.name)) {
		    out.println("if (element.checked) {");
		    out.println("  element.checked = PowerForms.toBoolean(PowerForms.equalValue(element.value, "+form.index+", '"+fieldList1.name+"'));");
		    out.println("  if (!element.checked) {");
		    out.println("    change = true;");
		    out.println("  }");
		    out.println("}");
		  }
		}
	      } else if (field instanceof HtmlSelect) {
		/*
		if ((constraint.fieldName != null) && constraint.fieldName.equals(fieldList0.name)) {
		  out.println("change |= element.filter("+regexp.getDfaName()+".filter);");
		} else {
		  String errorMessage = getErrorMessage("Invalid input in field "+field.name);
		  out.println("if (!element.equalValues("+Html.scriptQuote(value)+")) {");
		  out.println("  PowerForms.addMessage(messages, "+Html.scriptQuote(errorMessage)+");");
		  out.println("  accept = false;");
		  out.println("}");
		}
		*/
	      }
	    }
	  }
	  HtmlFieldList l = fieldList0;
	  fieldList0 = fieldList1;
	  fieldList1 = l;
	  if (count == 1) {
	    out.println("if (state == PowerForms.NO) {");
	  } else {
	    out.println("}");
	  }
	}
	if (!true) {
	  if ((fieldList0 != null) && (fieldList1 != null)) {
	    out.println("if (!PowerForms.toBoolean(PowerForms.equal("+form.index+", '"+fieldList0.name+"', "+form.index+", '"+fieldList1.name+"'))) {");
	    out.println("  PowerForms.addMessage(messages, "+Html.scriptQuote(getErrorMessage("fields `"+fieldList0.name+"' and `"+fieldList1.name+"' must be equal"))+");");
	    out.println("}");
	  }
	}
      }
    }
  }

  boolean evaluate(Map values) {
    return validate(values, null);
  }

  boolean validate(Map values, ErrorHandler handler) {
    if (value != null) {
      Collection c = HtmlForm.getValues(values, fieldNames[0]);
      if (c != null) {
	for (Iterator i = c.iterator(); i.hasNext(); ) {
	  String value = i.next().toString();
	  if (value.equals(this.value)) {
	    return true;
	  }
	}
      }
      if (handler != null) {
	handler.error(getErrorMessage(getPowerForms().validationError("powerforms.requireValue", fieldNames[0], value)));
      }
    } else {
      Collection c0 = HtmlForm.getValues(values, fieldNames[0]);
      Collection c1 = HtmlForm.getValues(values, fieldNames[1]);
      if ((c0 != null) && (c1 != null)) {
	for (Iterator i0 = c0.iterator(); i0.hasNext(); ) {
	  Object value = i0.next().toString();
	  /* PowerForms.log("value: "+value); */
	  for (Iterator i1 = c1.iterator(); i1.hasNext(); ) {
	    if (value.equals(i1.next().toString())) {
	      return true;
	    }
	  }
	  /*
	  // dk.brics.jwig.runwig.Arg doesn't implement "public boolean equal(Object o)"!
	  if (c1.contains(value)) {
	    return true;
	  }
	  */
	}
      }
      if (handler != null) {
	handler.error(getErrorMessage(getPowerForms().validationError("powerforms.unequalValues", fieldNames[0], fieldNames[1])));
      }
    }
    return false;
  }
}

class ExpressionField extends Expression {
  String name;

  ExpressionField(Attributes attributes) {
    init("field", attributes);
  }

  void check(boolean isTest) throws TransformException {
    name = attributes.requireValue("name");
  }

  boolean evaluate(Map values) {
    return true;
  }

  boolean validate(Map values, ErrorHandler handler) {
    return true;
  }
}

class ExpressionFunction extends Expression {
  private HtmlForm form;
  private String functionName;
  private HtmlFieldList[] arguments;
  private HtmlFieldList[] fields;
  String formName, fieldName;

  ExpressionFunction(Attributes attributes) {
    init("function", attributes);
  }

  void check(boolean isTest) throws TransformException {
    super.check(isTest);
    functionName = attributes.requireValue("name");
    formName = constraint.formName;
    if (isTest || (constraint.fieldName == null)) {
      fieldName = attributes.requireValue("field");
      constraint.addDependency(formName, fieldName);
    } else {
      fieldName = constraint.fieldName;
      attributes.disallowAttribute("field");
    }
    String[] argumentNames = attributes.getValues("argument", "arguments");
    /*
    if ((form != null) && (argumentNames != null)) {
      arguments = form.getFieldLists(this, argumentNames);
      for (int i = 0; i < arguments.length; i++) {
	arguments[i].addAffects(constraint);
      }
    }
    if (fields != null) {
      if (isTest) {
	for (int i = 0; i < fields.length; i++) {
	  fields[i].addAffects(constraint);
	}
      } else {
	for (int i = 0; i < fields.length; i++) {
	  fields[i].addConstraint(constraint);
	  for (int j = 0; j < fields[i].size(); j++) {
	    HtmlField field = fields[i].getField(j);
	    if (constraint.fields != null) {
	      field.getStatus();
	    }
	  }
	}
      }
    }
    */
  }

  void script(HtmlDocument document, PrintWriter out) {
    if (!isTest) {
      out.print("state = ");
    }
    out.print(functionName+"(");
    if (arguments != null) {
      for (int i = 0; i < arguments.length; i++) {
	if (i > 0) {
	  out.print(", ");
	}
	out.print("PowerForms.getValues("+form.index+", '"+arguments[i].name+"').getElements()");
      }
    }
    out.print(")");
    if (!isTest) {
      out.println(";");
      if (fields != null) {
	for (int i = 0; i < fields.length; i++) {
	  for (int j = 0; j < fields[i].size(); j++) {
	    HtmlField field = fields[i].getField(j);
	    if (constraint.fieldName != null) {
	      HtmlStatus status = field.getStatus();
	      if (status != null) {
		out.println(status.getScriptName()+".update(state);");
	      }
	      AutoComplete autocomplete = field.getAutoComplete();
	      if (autocomplete != null) {
		out.println(autocomplete.getScriptName()+".setCompletion(null);");
	      }
	    }
	  }
	}
      }
    }
  }

  boolean evaluate(Map values) {
    return true;
  }

  boolean validate(Map values, ErrorHandler handler) {
    return true;
  }
}

class ExpressionLessThan extends Expression {
  private static final int STRING = 0;
  private static final int INTEGER = 1;

  int type;
  String value;
  String formName;
  String[] fieldNames;

  ExpressionLessThan(Attributes attributes, PwfContent content) {
    init("less-than", attributes, content);
  }

  void check(boolean isTest) throws TransformException {
    super.check(isTest);
    formName = constraint.formName;
    if (isTest || (constraint.fieldName == null)) {
      String fieldName = attributes.getValue("field");
      if (fieldName != null) {
	fieldNames = new String[] { fieldName };
      } else {
	fieldNames = new String[content.size()];
	if (fieldNames.length == 0) {
	  reportError("missing fields for `"+tagName+"'");
	} else {
	  for (int i = 0; i < content.size(); i++) {
	    ExpressionField field = (ExpressionField)content.get(i);
	    fieldNames[i] = field.name;
	  }
	}
      }
      for (int i = 0; i < fieldNames.length; i++) {
	constraint.addDependency(formName, fieldNames[i]);
      }
    } else {
      String fieldName = attributes.getValue("field");
      if (fieldName != null) {
	fieldNames = new String[] { constraint.fieldName, fieldName };
	constraint.addDependency(formName, fieldName);
      } else {
	fieldNames = new String[1+content.size()];
	fieldNames[0] = constraint.fieldName;
	for (int i = 0; i < content.size(); i++) {
	  ExpressionField field = (ExpressionField)content.get(i);
	  fieldNames[i+1] = field.name;
	}
      }
    }
    if (fieldNames.length > 2) {
      reportError("only two fields allowed for `"+tagName+"'");
    } else if (fieldNames.length == 1) {
      value = attributes.requireValue("value");
    } else {
      attributes.disallowAttribute("value");
    }
    String s = attributes.getValue("type", "string");
    if (s.equals("string")) {
      type = STRING;
    } else if (s.equals("integer")) {
      type = INTEGER;
    } else {
      reportError("unknown type `"+s+"' for '"+tagName+"'");
    }
  }

  void script(HtmlDocument document, PrintWriter out) {
    HtmlForm form = document.getForm(formName);
    if (isTest) {
      if (fieldNames.length == 1) {
	HtmlFieldList fieldList = (form == null ? null : form.getFieldList(fieldNames[0]));
	if (fieldList == null) {
	  out.print("PowerForms.YES");
	} else {
	  out.print("PowerForms.lessThan("+form.index+", '"+fieldList.name+"', "+Html.scriptQuote(value)+", PowerForms."+(type == STRING ? "STRING" : "INTEGER")+")");
	}
      } else {
	HtmlFieldList fieldList0 = form.getFieldList(fieldNames[0]);
	HtmlFieldList fieldList1 = form.getFieldList(fieldNames[1]);
	if ((fieldList0 == null) || (fieldList1 == null)) {
	  out.print("PowerForms.YES");
	} else {
	  out.print("PowerForms.lessThan("+form.index+", '"+fieldList0.name+"', "+form.index+", '"+fieldList1.name+"', PowerForms."+(type == STRING ? "STRING" : "INTEGER")+")");
	}
      }
    } else {
      if (fieldNames.length == 1) {
	HtmlFieldList fieldList = (form == null ? null : form.getFieldList(fieldNames[0]));
	if (fieldList != null) {
	  for (int i = 0; i < fieldList.size(); i++) {
	    HtmlField field = fieldList.getField(i);
	    out.println("element = document.forms["+form.index+"].elements["+field.index+"];");
	    if (field instanceof HtmlInput) {
	      HtmlInput input = (HtmlInput)field;
	      if ((input.type == HtmlInput.TEXT) || (input.type == HtmlInput.PASSWORD)) {
		out.println("state = PowerForms.lessThanValue(element.value, "+Html.scriptQuote(value)+");");
		out.println("switch (state) {");
		out.println("case PowerForms.NO:");
		/* out.println("  PowerForms.addMessage(messages, "+Html.scriptQuote(getErrorMessage("field `"+fieldNames[0]+"' must be less than `"+value+"'"))+");"); */
		/* out.println("  accept = false;"); */
		/* out.println("  break;"); */
		out.println("ca"+"se PowerForms.MAYBE:");
		out.println("  PowerForms.addMessage(messages, "+Html.scriptQuote(getErrorMessage("field `"+fieldNames[0]+"' must be less than `"+value+"'"))+");");
		out.println("  accept = false;");
		out.println("  break;");
		out.println("}");
		if (constraint.fieldName != null) {
		  HtmlStatus status = field.getStatus();
		  if (status != null) {
		    out.println(status.getScriptName()+".update(state);");
		  }
		  AutoComplete autocomplete = field.getAutoComplete();
		  if (autocomplete != null) {
		    out.println("if (state == PowerForms.MAYBE) {");
		    out.print(autocomplete.getScriptName()+".setCompletion(PowerForms.lessThanComplete");
		    switch (autocomplete.type) {
		    case AutoComplete.FIRST:
		      out.print("First(element.value, "+Html.scriptQuote(value)+")");
		      break;
		    case AutoComplete.PREFIX:
		    case AutoComplete.AGGRESSIVE:
		      out.print("Unique(element.value, "+Html.scriptQuote(value)+", true)");
		      break;
		    default:
		      out.print("Unique(element.value, "+Html.scriptQuote(value)+", false)");
		      break;
		    }
		    out.println(");");
		    if (autocomplete.type == AutoComplete.AGGRESSIVE) {
		      out.println(autocomplete.getScriptName()+".complete();");
		    }
		    out.println("} else {");
		    out.println(autocomplete.getScriptName()+".setCompletion(null);");
		    out.println("}");
		  }
		}
	      } else if ((input.type == HtmlInput.RADIO) || (input.type == HtmlInput.CHECKBOX)) {
		if (constraint.fieldName != null) {
		  out.println("if (element.checked) {");
		  out.println("  element.checked = PowerForms.toBoolean(PowerForms.lessThan(element.value, "+Html.scriptQuote(value)+"));");
		  out.println("  if (!element.checked) {");
		  out.println("    change = true;");
		  out.println("  }");
		  out.println("}");
		}
	      }
	    } else if (field instanceof HtmlSelect) {
	      /*
		if (constraint.fieldName != null) {
		out.println("change |= element.filter("+regexp.getDfaName()+".filter);");
		} else {
		String errorMessage = getErrorMessage("Invalid input in field "+field.name);
		out.println("if (!element.equalValues("+Html.scriptQuote(value)+")) {");
		out.println("  PowerForms.addMessage(messages, "+Html.scriptQuote(errorMessage)+");");
		out.println("  accept = false;");
		out.println("}");
		}
	      */
	    }
	  }
	}
      } else {
	HtmlFieldList fieldList0 = form.getFieldList(fieldNames[0]);
	HtmlFieldList fieldList1 = form.getFieldList(fieldNames[1]);
	int count = 0;
	while (count++ < 2) {
	  if ((fieldList0 != null) && (fieldList1 != null)) {
	    for (int i = 0; i < fieldList0.size(); i++) {
	      HtmlField field = fieldList0.getField(i);
	      out.println("element = document.forms["+form.index+"].elements["+field.index+"];");
	      if (field instanceof HtmlInput) {
		HtmlInput input = (HtmlInput)field;
		if ((input.type == HtmlInput.TEXT) || (input.type == HtmlInput.PASSWORD)) {
		  out.println("state = PowerForms.lessThanValue(element.value, "+form.index+", "+Html.scriptQuote(fieldList1.name)+");");
		  out.println("switch (state) {");
		  out.println("case PowerForms.NO:");
		  /* out.println("  PowerForms.addMessage(messages, "+Html.scriptQuote(getErrorMessage("field `"+fieldNames[0]+"' must be less than `"+value+"'"))+");"); */
		  /* out.println("  accept = false;"); */
		  /* out.println("  break;"); */
		  out.println("ca"+"se PowerForms.MAYBE:");
		  out.println("  PowerForms.addMessage(messages, "+Html.scriptQuote(getErrorMessage("field `"+fieldNames[0]+"' must be less than `"+value+"'"))+");");
		  out.println("  accept = false;");
		  out.println("  break;");
		  out.println("}");
		  if ((constraint.fieldName != null) && constraint.fieldName.equals(fieldList0.name)) {
		    HtmlStatus status = field.getStatus();
		    if (status != null) {
		      out.println(status.getScriptName()+".update(state);");
		    }
		    AutoComplete autocomplete = field.getAutoComplete();
		    if (autocomplete != null) {
		      out.println("if (state == PowerForms.MAYBE) {");
		      out.print(autocomplete.getScriptName()+".setCompletion(PowerForms.lessThanComplete");
		      switch (autocomplete.type) {
		      case AutoComplete.FIRST:
			out.print("First(element.value, "+form.index+", "+Html.scriptQuote(fieldList1.name)+")");
			break;
		      case AutoComplete.PREFIX:
		      case AutoComplete.AGGRESSIVE:
			out.print("Unique(element.value, "+form.index+", "+Html.scriptQuote(fieldList1.name)+", true)");
			break;
		      default:
			out.print("Unique(element.value, "+form.index+", "+Html.scriptQuote(fieldList1.name)+", false)");
			break;
		      }
		      out.println(");");
		      if (autocomplete.type == AutoComplete.AGGRESSIVE) {
			out.println(autocomplete.getScriptName()+".complete();");
		      }
		      out.println("} else {");
		      out.println(autocomplete.getScriptName()+".setCompletion(null);");
		      out.println("}");
		    }
		  }
		} else if ((input.type == HtmlInput.RADIO) || (input.type == HtmlInput.CHECKBOX)) {
		  if ((constraint.fieldName != null) && constraint.fieldName.equals(fieldList0.name)) {
		    out.println("if (element.checked) {");
		    out.println("  element.checked = PowerForms.toBoolean(PowerForms.lessThanValue(element.value, "+form.index+", '"+fieldList1.name+"'));");
		    out.println("  if (!element.checked) {");
		    out.println("    change = true;");
		    out.println("  }");
		    out.println("}");
		  }
		}
	      } else if (field instanceof HtmlSelect) {
		/*
		  if ((constraint.fieldName != null) && constraint.fieldName.equals(fieldList0.name)) {
		  out.println("change |= element.filter("+regexp.getDfaName()+".filter);");
		  } else {
		  String errorMessage = getErrorMessage("Invalid input in field "+field.name);
		  out.println("if (!element.equalValues("+Html.scriptQuote(value)+")) {");
		  out.println("  PowerForms.addMessage(messages, "+Html.scriptQuote(errorMessage)+");");
		  out.println("  accept = false;");
		  out.println("}");
		  }
		*/
	      }
	    }
	  }
	  HtmlFieldList l = fieldList0;
	  fieldList0 = fieldList1;
	  fieldList1 = l;
	}
      }
    }
  }

  boolean evaluate(Map values) {
    return validate(values, null);
  }

  boolean validate(Map values, ErrorHandler handler) {
    if (value != null) {
      Collection c = HtmlForm.getValues(values, fieldNames[0]);
      if (c != null) {
	for (Iterator i = c.iterator(); i.hasNext(); ) {
	  String value = i.next().toString();
	  if (lessThan(value, this.value, type)) {
	    return true;
	  }
	}
      }
      if (handler != null) {
	handler.error(getErrorMessage(getPowerForms().validationError("powerforms.notLessThanValue", fieldNames[0], value)));
      }
    } else {
      Collection c0 = HtmlForm.getValues(values, fieldNames[0]);
      Collection c1 = HtmlForm.getValues(values, fieldNames[1]);
      if ((c0 != null) && (c1 != null)) {
	for (Iterator i0 = c0.iterator(); i0.hasNext(); ) {
	  Object value = i0.next().toString();
	  for (Iterator i1 = c1.iterator(); i1.hasNext(); ) {
	    if (lessThan(value, i1.next().toString(), type)) {
	      return true;
	    }
	  }
	}
      }
      if (handler != null) {
	handler.error(getErrorMessage(getPowerForms().validationError("powerforms.notLessThanFields", fieldNames[0], fieldNames[1])));
      }
    }
    return false;
  }

  private static boolean lessThan(Object a, Object b, int type) {
    switch (type) {
    case STRING:
      return (a.toString().compareTo(b.toString()) < 0);
    case INTEGER:
      Integer i = null;
      Integer j = null;
      try {
	i = Integer.valueOf(a.toString().trim());
      } catch (NumberFormatException e) {
	i = null;
      }
      try {
	j = Integer.valueOf(b.toString().trim());
      } catch (NumberFormatException e) {
	j = null;
      }
      if ((i != null) && (j != null)) {
	return (i.compareTo(j) < 0);
      }
      return false;
    }
    return false;
  }
}

class ExpressionAnd extends Expression {
  ExpressionAnd(PwfContent expressions) {
    init("and", expressions);
  }

  void check(boolean isTest) throws TransformException {
    super.check(isTest);
    if (!isTest) {
      reportError("`"+tagName+"' only allowed as boolean expression");
    }
  }

  void script(HtmlDocument document, PrintWriter out) {
    out.print("PowerForms.and(");
    for (int i = 0; i < content.size(); i++) {
      if (i > 0) {
	out.print(", ");
      }
      content.get(i).script(document, out);
    }
    out.print(")");
  }

  boolean evaluate(Map values) {
    for (int i = 0; i < content.size(); i++) {
      Expression e = (Expression)content.get(i);
      if (!e.evaluate(values)) {
	return false;
      }
    }
    return true;
  }

  boolean validate(Map values, ErrorHandler handler) {
    return true;
  }
}

class ExpressionOr extends Expression {
  ExpressionOr(PwfContent expressions) {
    init("or", expressions);
  }

  void check(boolean isTest) throws TransformException {
    super.check(isTest);
    if (!isTest) {
      reportError("`"+tagName+"' only allowed as boolean expression");
    }
  }

  void script(HtmlDocument document, PrintWriter out) {
    out.print("PowerForms.or(");
    for (int i = 0; i < content.size(); i++) {
      if (i > 0) {
	out.print(", ");
      }
      content.get(i).script(document, out);
    }
    out.print(")");
  }

  boolean evaluate(Map values) {
    for (int i = 0; i < content.size(); i++) {
      Expression e = (Expression)content.get(i);
      if (e.evaluate(values)) {
	return true;
      }
    }
    return false;
  }

  boolean validate(Map values, ErrorHandler handler) {
    return true;
  }
}

class ExpressionNot extends Expression {
  ExpressionNot(PwfContent expressions) {
    init("not", expressions);
  }

  void check(boolean isTest) throws TransformException {
    super.check(isTest);
    if (!isTest) {
      reportError("`"+tagName+"' only allowed as boolean expression");
    }
  }

  void script(HtmlDocument document, PrintWriter out) {
    out.print("PowerForms.not(");
    for (int i = 0; i < content.size(); i++) {
      if (i > 0) {
	out.print(", ");
      }
      content.get(i).script(document, out);
    }
    out.print(")");
  }

  boolean evaluate(Map values) {
    for (int i = 0; i < content.size(); i++) {
      Expression e = (Expression)content.get(i);
      if (e.evaluate(values)) {
	return false;
      }
    }
    return true;
  }

  boolean validate(Map values, ErrorHandler handler) {
    return true;
  }
}
