/*
 * PowerForms
 * Copyright (C) 1999-2002 BRICS.
 *
 * Author: Mikkel Ricky <ricky@brics.dk>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the  Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 * USA.
 *
 * Home page: http://www.brics.dk/bigwig/powerforms/
 */

package dk.brics.powerforms;

import dk.brics.automaton.*;
import java.io.*;
import java.net.*;
import java.util.*;
import java.text.*;

abstract class Html {
  static String scriptQuote(String s) {
    if (s == null) {
      return null;
    }
    if (s.indexOf('\'') < 0) {
      return "'"+s+"'";
    } else if (s.indexOf('"') < 0) {
      return '"'+s+'"';
    } else {
      StringBuffer sb = new StringBuffer();
      sb.append("'");
      for (int i = 0; i < s.length(); i++) {
	if (s.charAt(i) == '\'') {
	  sb.append("\\");
	}
	sb.append(s.charAt(i));
      }
      sb.append("'");
      return sb.toString();
    }
  }
}


/* class HtmlSomething extends PwfNode { */
class HtmlSomething extends HtmlElement {
  private String text;

  HtmlSomething(String text) {
    this.text = text;
  }

  String getText() {
    return text;
  }

  void append(HtmlSomething something) {
    this.text += something.text;
  }

  void check() throws TransformException {}

  void script(HtmlDocument document, PrintWriter out) {}
  void scriptDefine(HtmlDocument document, PrintWriter out) {}
  void scriptInit(HtmlDocument document, PrintWriter out) {}

  void code(PrintWriter out) {
    int index = text.toLowerCase().indexOf("</head");
    if (index > -1) {
      out.print(text.substring(0, index));
      getDocument().script(out);
      out.print(text.substring(index));
    } else {
      out.print(text);
    }
  }

  void print(PrintWriter out) {
    out.print(text);
  }

  public String toString() {
    return "[Text node: "+(text.length() > 20 ? text.substring(0, 16)+" ...": text)+"]";
  }
}


class HtmlForm extends HtmlElement {
  final static Set invalidFieldNames = new HashSet();
  static {
    String[] names = { "reset", "length", "submit" };
    for (int i = 0; i < names.length; i++) {
      invalidFieldNames.add(names[i]);
    }
  }

  private int elementIndex = 0;
  private int constraintIndex = 0;

  List fields;
  int index;
  boolean hasConstraints, hasReset, hasResult, hasSelect, hasSubmitInput, hasSubmitElement;
  boolean hasJWIGSubmitAnchor = false;
  Set submitNames;
  Map submitFields;

  HtmlForm(Attributes attributes, HtmlContent content) {
    init("form", attributes, content);
  }

  private void addSubmitInfo(String name, String value, boolean ignoreconstraints) {
    StringPair key = new StringPair(name, value);
    submitNames.add(name);
    if (ignoreconstraints) {
      submitFields.put(key, Boolean.TRUE);
    } else {
      Boolean ignore = (Boolean)submitFields.get(key);
      submitFields.put(key, ignore == null ? Boolean.FALSE : ignore);
    }
  }

  void addSubmit(HtmlSubmit submit) {
    hasSubmitElement = true;
    addSubmitInfo(submit.name, submit.value, submit.ignoreconstraints);
    if ("button".equals(submit.submittype)) {
      elementIndex++;
    }
  }

  void addJWIGSubmitAnchor(String value, boolean ignoreconstraints) {
    hasSubmitElement = true;
    hasJWIGSubmitAnchor = true;
    addSubmitInfo(value, value, ignoreconstraints);
  }

  void addField(HtmlField field) throws TransformException {
    if ((field.name != null) && invalidFieldNames.contains(field.name)) {
      field.reportError("invalid name `"+field.name+"' for "+field.getTagName());
    } else {
      if (field.name != null) {
	HtmlFieldList fieldList = getFieldList(field.name);
	if (fieldList == null) {
	  fieldList = new HtmlFieldList(this, field.name);
	  fields.add(fieldList);
	}
	fieldList.addField(field);
      }
      field.index = elementIndex++;
      if (field instanceof HtmlInput) {
	HtmlInput input = (HtmlInput)field;
	if (input.type == HtmlInput.RESET) {
	  hasReset = true;
	}
	if (input.type == HtmlInput.SUBMIT) {
	  hasSubmitInput = true;
	  addSubmitInfo(input.name, input.value, input.ignoreconstraints);
	  HtmlFieldList fieldList = getFieldList(input.name);
	  if (fieldList != null) {
	    for (int i = 0; i < fieldList.size(); i++) {
	      HtmlField f = fieldList.getField(i);
	      if (f != field) {
		if (f.type != HtmlField.SUBMIT) {
		  f.reportError("an input field cannot have the same name (`"+input.name+"') as a submit field");
		}
	      }
	    }
	  }
	} else if (submitNames.contains(input.name)) {
	  field.reportError("an input field cannot have the same name (`"+input.name+"') as a submit field");
	}
      } else if (field instanceof HtmlButton) {
	HtmlButton button = (HtmlButton)field;
	if (button.type == HtmlInput.RESET) {
	  hasReset = true;
	}
	if (button.type == HtmlInput.SUBMIT) {
	  hasSubmitElement = true;
	  addSubmitInfo(button.name, button.value, button.ignoreconstraints);
	  HtmlFieldList fieldList = getFieldList(button.name);
	  if (fieldList != null) {
	    for (int i = 0; i < fieldList.size(); i++) {
	      HtmlField f = fieldList.getField(i);
	      if (f != field) {
		if (f.type != HtmlField.SUBMIT) {
		  f.reportError("an input field cannot have the same name (`"+button.name+"') as a submit field");
		}
	      }
	    }
	  }
	}
      } else if (field instanceof HtmlResult) {
	hasResult = true;
      } else if (field instanceof HtmlSelect) {
	hasSelect = true;
      } else if (field instanceof HtmlTextArea) {
      }
    }
  }

  HtmlFieldList getFieldList(String name) {
    for (int i = 0; i < fields.size(); i++) {
      HtmlFieldList fieldList = (HtmlFieldList)fields.get(i);
      if (fieldList.name.equals(name)) {
	return fieldList;
      }
    }
    return null;
  }

  HtmlFieldList[] getFieldLists(PwfElement element, String[] fieldNames) {
    HtmlFieldList[] list = new HtmlFieldList[fieldNames.length];
    int size = 0;
    for (int i = 0; i < fieldNames.length; i++) {
      HtmlFieldList fl = getFieldList(fieldNames[i]);
      if (fl == null) {
	/* element.reportError("no field named `"+fieldNames[i]+"' on form"); */
      } else {
	list[size++] = fl;
      }
    }
    if (size < list.length) {
      HtmlFieldList[] tmp = list;
      list = new HtmlFieldList[size];
      System.arraycopy(tmp, 0, list, 0, list.length);
    }
    return list;
  }

  void check() throws TransformException {
    fields = new ArrayList();
    submitNames = new HashSet();
    submitFields = new HashMap();
    index = getDocument().getFormNumber();
    if (index > 0) {
      attributes.addID();
    }
    getDocument().addForm(this);
    if (isNested()) {
      reportError("nested forms not allowed");
    } else {
      super.check();
    }
  }

  private boolean resetCode = false;
  private boolean submitCode = false;
  private boolean updateCode = false;

  String getUpdateFunctionName() {
    return updateCode ? "pwf_update_form_"+index : null;
  }

  void defineUpdateFunction(PrintWriter out) {
    out.println();
    out.println("function "+getUpdateFunctionName()+"() {");
    /* out.println("PowerForms.DEBUG('"+getUpdateFunctionName()+"()');"); */
    out.print("  return "+getDocument().getUpdateFunctionName()+"(new Array(");
    ListSet constraints = getDocument().getConstraints(this);
    for (Iterator i = constraints.iterator(); i.hasNext(); ) {
      ConstraintApply c = (ConstraintApply)i.next();
      out.print(c.getScriptName());
      if (i.hasNext()) {
	out.print(", ");
      }
    }
    out.println("));");
    out.println("}");
  }

  String getResetFunctionName() {
    return resetCode ? "pwf_reset_form_"+index : null;
  }

  void defineResetFunction(PrintWriter out) {
    out.println();
    out.println("function "+getResetFunctionName()+"() {");
    /* out.println("PowerForms.DEBUG('"+getResetFunctionName()+"()');"); */
    out.println("  document.forms["+index+"].submitted = false;");
    out.println("  document.forms["+index+"].reset();");
    if (submitCode) {
      out.println("  document.forms["+index+"].elements["+elementIndex+"].name = null;");
      out.println("  document.forms["+index+"].elements["+elementIndex+"].value = '';");
    }
    out.println("  "+getUpdateFunctionName()+"();");
    out.println("}");
  }

  String getSubmitFunctionName() {
    return submitCode ? "pwf_submit_form_"+index : null;
  }

  void defineSubmitFunction(PrintWriter out) {
    out.println();
    out.println("function "+getSubmitFunctionName()+"(name, value, ignoreconstraints) {");
    /* out.println("PowerForms.DEBUG('"+getSubmitFunctionName()+"()');"); */
    out.println("if (document.forms["+index+"].submitted) {");

    /* out.println("  confirm('Form already submitted!');"); */

    out.println("  return false;");
    out.println("}");
    if (updateCode) {
      out.println("  pwf_event = null;");
      out.println("  if (!ignoreconstraints) {");
      out.println("    var info = "+getUpdateFunctionName()+"();");
      out.println("    if (!info[0]) {");
      out.println("      var messages = info[2];");
      /*
	out.println("      var msg = '';");
	out.println("      for (var i = 0; i < messages.length; i++) {");
	out.println("        msg += messages[i]+'\\n';");
	out.println("      }");
	out.println("      confirm(messages.length+' error'+(messages.length == 1 ? '' : 's')+' on form:\\n'+msg);");
      */
      String errortitle = getPowerForms().getErrorTitle();
      out.println("      confirm("+(errortitle != null ? Html.scriptQuote(errortitle) : "messages.length+' error'+(messages.length == 1 ? '' : 's')+' on form:'")+"+'\\n'+messages.join('\\n'));");
      out.println("      return false;");
      out.println("    }");
      out.println("  }");
    }
    out.println("  document.forms["+index+"].submitted = true;");
    out.println("  document.forms["+index+"].onsubmit = null;");
    if (hasJWIGSubmitAnchor) {
      out.println("  if ((name == '$submit') && (typeof(submitbutton) == 'function')) {");
      /* out.println("    document.forms['"+attributes.getValue("name")+"']['$submit'].value = value;"); */
      out.println("    submitbutton('"+attributes.getValue("name")+"', value);");
      out.println("    return true;");
      out.println("  }");
    }
    out.println("  if (name != null) {");
    out.println("    document.forms["+index+"].elements["+elementIndex+"].name = name;");
    out.println("    document.forms["+index+"].elements["+elementIndex+"].value = (value == null ? '' : value);");
    out.println("  }");
    out.println("  document.forms["+index+"].submit();");
    out.println("  return true;");
    out.println("}");
  }

  void scriptDefine(HtmlDocument document, PrintWriter out) {
    hasConstraints = getDocument().hasConstraints(this);
    updateCode = hasConstraints;
    resetCode = hasReset && hasConstraints;
    submitCode = hasConstraints || hasSubmitElement;
    /* submitCode = hasSubmitElement || (hasSubmitInput && updateCode); */

    if (updateCode) {
      defineUpdateFunction(out);
    }
    if (resetCode) {
      defineResetFunction(out);
    }
    if (submitCode) {
      defineSubmitFunction(out);
    }
    super.scriptDefine(document, out);
  }

  void code(PrintWriter out) {
    if (index == 0) {
      getDocument().script(out);
    }

    /*
    out.println("<!--");
    out.println("  fields: "+fields);
    out.println("  submitNames: "+submitNames);
    out.println("  submitFields: "+submitFields);
    out.println("-->");
    */

    out.print("<"+tagName);
    if (submitCode) {
      String onsubmit = attributes.removeHtml("onsubmit");
      if (onsubmit != null) {
	attributes.add("onsubmit", onsubmit +"; return "+getSubmitFunctionName()+"(null, null, false);");
      } else {
	attributes.add("onsubmit", "return "+getSubmitFunctionName()+"(null, null, false);");
      }
    }
    attributes.print(out);
    out.print(">");
    content.code(out);
    /* if ((id != null) && hasConstraints) { */
    if (id != null) {
      String formName = getPowerForms().getFormValidateName();
      if (formName != null) {
	out.print("<input type='hidden' name='"+formName+"' value='"+id+"' />");
      }
    }
    if (submitCode) {
      out.print("<input type='hidden' name='' />");
    }
    out.print("</"+tagName+">");
    if (getDocument().isLastForm(index)) {
      String initFunctionName = getDocument().getInitFunctionName();
      if (initFunctionName != null) {
	out.print("<script type='text/javascript'>"+initFunctionName+"();</script>");
      }
    }
  }

  boolean validate(Map values, ErrorHandler handler, boolean jwig) {
    boolean ignoreconstraints = false;
    if (jwig) {
      String submitName = "submit";
      Collection c = HtmlForm.getValues(values, submitName);
      if (c != null) {
	String submitValue = null;
	for (Iterator i = c.iterator(); i.hasNext(); ) {
	  submitValue = i.next().toString();
	  break;
	}
	if (submitValue != null) {
	  for (Iterator i = submitFields.keySet().iterator(); i.hasNext(); ) {
	    StringPair p = (StringPair)i.next();
	    if (submitValue.equals(p.first)) {
	      Boolean ignore = (Boolean)submitFields.get(p);
	      if ((ignore != null) && ignore.booleanValue()) {
		ignoreconstraints = true;
	      }
	    }
	  }
	}
      }
    } else {
      for (Iterator i = submitNames.iterator(); i.hasNext(); ) {
	String name = i.next().toString();
	Collection c = HtmlForm.getValues(values, name);
	if (c != null) {
	  for (Iterator j = c.iterator(); j.hasNext(); ) {
	    String value = j.next().toString();
	    Boolean ignore = (Boolean)submitFields.get(new StringPair(name, value));
	    if ((ignore != null) && ignore.booleanValue()) {
	      ignoreconstraints = true;
	    }
	  }
	}
      }
    }
    if (ignoreconstraints) {
      return true;
    } else {
      boolean errors = false;
      // Form constraints ...
      ListSet constraints = getPowerForms().getConstraints(id, null);
      if (constraints != null) {
	for (Iterator it = constraints.iterator(); it.hasNext(); ) {
	  ConstraintApply constraint = (ConstraintApply)it.next();
	  if (!constraint.validate(values, handler)) {
	    errors = true;
	  }
	}
      }
      // Field constraints ...
      for (int i = 0; i < fields.size(); i++) {
	HtmlFieldList fieldList = (HtmlFieldList)fields.get(i);
	if (!jwig || ((fieldList.name != null)
		      && !fieldList.name.startsWith("$")
		      && !fieldList.name.startsWith("*")
		      && !fieldList.name.startsWith("#"))) {
	  Collection c = HtmlForm.getValues(values, fieldList.name);
	  if (c == null) {
	    if (0 < fieldList.lowerBound) {
	      handler.error(getPowerForms().validationError("powerforms.tooFewValues", fieldList.name));
	      errors = true;
	    }
	  } else if (c.size() < fieldList.lowerBound) {
	    handler.error(getPowerForms().validationError("powerforms.tooFewValues", fieldList.name));
	    errors = true;
	  } else if (fieldList.upperBound < c.size()) {
	    handler.error(getPowerForms().validationError("powerforms.tooManyValues", fieldList.name));
	    errors = true;
	  } else {
	    constraints = getPowerForms().getConstraints(id, fieldList.name);
	    if (constraints != null) {
	      for (Iterator it = constraints.iterator(); it.hasNext(); ) {
		ConstraintApply constraint = (ConstraintApply)it.next();
		if (!constraint.validate(values, handler)) {
		  errors = true;
		}
	      }
	    }
	  }
	}
      }
      return !errors;
    }
  }

  static Collection getValues(Map values, Object key) {
    Collection c = (Collection)values.get(key);
    if (c == null) {
      c = (Collection)values.get("#"+key);
    }
    return c;
  }

  public String toString() {
    CharArrayWriter chars = new CharArrayWriter();
    PrintWriter out = new PrintWriter(chars);
    out.println("Form ("+id+")");
    for (int j = 0; j < fields.size(); j++) {
      out.println("  "+fields.get(j));
    }
    out.println("  submitNames: "+submitNames);
    out.println("  submitFields: "+submitFields);
    out.close();
    return chars.toString();
  }
}


class HtmlInput extends HtmlField {
  boolean ignoreconstraints = false;
  AutoComplete autocomplete;

  HtmlInput(Attributes attributes) {
    init("input", attributes);
  }

  void check() throws TransformException {
    String t = attributes.getValueHtml("type");
    if (t == null) {
      type = TEXT;
    } else {
      t = t.toLowerCase().intern();
      if (t == "button") {
	type = BUTTON;
      } else if (t == "checkbox") {
	type = CHECKBOX;
      } else if (t == "file") {
	type = FILE;
      } else if (t == "hidden") {
	type = HIDDEN;
      } else if (t == "image") {
	type = IMAGE;
      } else if (t == "password") {
	type = PASSWORD;
      } else if (t == "radio") {
	type = RADIO;
      } else if (t == "reset") {
	type = RESET;
      } else if (t == "submit") {
	type = SUBMIT;
      } else if (t == "text") {
	type = TEXT;
      } else {
	reportError("unknown type `"+t+"' for input");
      }
    }
    if (type == HtmlField.SUBMIT) {
      ignoreconstraints = "yes".equalsIgnoreCase(attributes.remove("ignoreconstraints"));
    }
    /*
      String s = "yes".equalsIgnoreCase(attributes.remove("autocomplete"));
    */
    super.check();
    /* if ((name == null) && !(type.equals("button") || type.equals("submit") || type.equals("reset"))) { */
    if ((name == null) && (type != BUTTON) && (type != RESET)) {
      attributes.requireValue("name");
    }
    /*
    // FIXME
    if (type.equals("image")) {
    form.addField(name+".x", this);
    form.addField(name+".y", this);
    } else {
    */
  }

  void scriptDefine(HtmlDocument document, PrintWriter out) {
    if (constraints != null) {
      getStatus();
      if (status != null) {
	status.scriptDefine(document, out);
      }
      getAutoComplete();
      if (autocomplete != null) {
	autocomplete.scriptDefine(document, out);
      }
    }
  }

  void scriptInit(HtmlDocument document, PrintWriter out) {
    if ((constraints != null) && (status != null)) {
      status.scriptInit(document, out);
    }
    if (form != null) {
      String updateFunction = getDocument().getUpdateFunctionName(form.id, name);
      if ((updateFunction != null) || (initValue != null)) {
	out.println("  element = document.forms["+form.index+"].elements["+index+"];");
	if (id != null) {
	  out.println("  PowerForms.addElementById('"+id+"', element);");
	}
	if (updateFunction != null) {
	  if ((type == TEXT) || (type == PASSWORD)) {
	    out.println("  if (element.addEventListener) {");
	    out.println("    element.addEventListener('onkeyup', "+updateFunction+", true);");
	    out.println("    element.addEventListener('onchange', "+updateFunction+", true);");
	    out.println("  } else {");
	    out.println("    element.onkeyup = isAcademicBrowser ? function(e) { this.blur(); this.focus(); "+updateFunction+"; } : "+updateFunction+";");
	    out.println("    element.onchange = "+updateFunction+";");
	    out.println("  }");
	  } else if ((type == RADIO) || (type == CHECKBOX)) {
	    out.println("  if (element.addEventListener) {");
	    out.println("    element.addEventListener('onclick', "+updateFunction+", true);");
	    out.println("  } else {");
	    out.println("    element.onclick = "+updateFunction+";");
	    out.println("  }");
	    if (true || hasConstraints()) {
	      out.println("  initCheck(element);");
	    }
	  }
	}
	if (initValue != null) {
	  if (((type == TEXT) || (type == PASSWORD)) && (value == null)) {
	    out.println("  element.value = "+Html.scriptQuote(initValue)+";");
	  } else if (((type == RADIO) || (type == CHECKBOX))
		     && initValue.equals(value) && (attributes.getValueHtml("checked") == null)) {
	    out.println("  element.checked = true;");
	  }
	}
      }
      if ((constraints != null) && (autocomplete != null)) {
	autocomplete.scriptInit(document, out);
      }
    }
  }

  void code(PrintWriter out) {
    if (autocomplete != null) {
      autocomplete.code(out);
    }
    out.print("<"+tagName);
    if ((type == SUBMIT) || (type == IMAGE)) {
      String submitFunctionName = (form == null ? null : form.getSubmitFunctionName());
      if (submitFunctionName != null) {
	out.print(" onclick='return "+submitFunctionName+"(");
	out.print(name == null ? "null" : "\""+name+"\"");
	/* out.print("null"); */
	out.print(", ");
	out.print(value == null ? "null" : "\""+value+"\"");
	/* out.print("null"); */
	out.print(", ");
	out.print(ignoreconstraints);
	out.print(");'");
      }
    } else if (type == RESET) {
      String resetFunctionName = (form == null ? null : form.getResetFunctionName());
      if (resetFunctionName != null) {
	out.print(" onclick='return "+resetFunctionName+"();'");
      }
    }
    if (initValue != null) {
      if (((type == TEXT) || (type == PASSWORD)) && (value == null)) {
	out.print(" value="+Html.scriptQuote(initValue));
      } else if (((type == RADIO) || (type == CHECKBOX))
		 && initValue.equals(value) && (attributes.getValueHtml("checked") == null)) {
	out.print(" checked='checked'");
      }
    }
    attributes.code(out);
    if (form != null) {
      String updateFunctionName = getDocument().getUpdateFunctionName(form.id, name);
      if (updateFunctionName != null) {
	if ((type == TEXT) || (type == PASSWORD)) {
	  out.print(" onkeyup='"+updateFunctionName+"();'");
	  out.print(" onchange='"+updateFunctionName+"();'");
	} else if ((type == RADIO) || (type == CHECKBOX)) {
	  out.print(" onclick='"+updateFunctionName+"();'");
	}
      }
    }
    out.print(" />");
    if ((status != null) && status.attributes.hasValue("default-status")) {
      status.code(out);
    }
  }

  HtmlStatus getStatus() {
    if ((status == null) && ((type == TEXT) || (type == PASSWORD))) {
      status = getDocument().getDefaultStatus(this);
    }
    return super.getStatus();
  }

  AutoComplete getAutoComplete() {
    if (autocomplete == null) {
      autocomplete = (type == TEXT ? getDocument().getAutoComplete(this) : null);
    }
    return autocomplete;
  }

  public String toString() {
    String t = null;
    switch (type) {
    case BUTTON:
      t = "button";
      break;
    case CHECKBOX:
      t = "checkbox";
      break;
    case FILE:
      t = "file";
      break;
    case HIDDEN:
      t = "hidden";
      break;
    case IMAGE:
      t = "image";
      break;
    case PASSWORD:
      t = "password";
      break;
    case RADIO:
      t = "radio";
      break;
    case RESET:
      t = "reset";
      break;
    case SUBMIT:
      t = "submit";
      break;
    case TEXT:
      t = "text";
      break;
    }
    return "["+t+" "+name+"]";
  }
}


class HtmlButton extends HtmlField {
  boolean ignoreconstraints = false;

  HtmlButton(Attributes attributes, HtmlContent content) {
    init("button", attributes, content);
  }

  void check() throws TransformException {
    String t = attributes.getValueHtml("type");
    if (t == null) {
      type = SUBMIT;
    } else {
      t = t.toLowerCase().intern();
      if (t == "button") {
	type = BUTTON;
      } else if (t == "reset") {
	type = RESET;
      } else if (t == "submit") {
	type = SUBMIT;
      } else {
	reportError("unknown type `"+t+"' for button");
      }
    }
    if (type == SUBMIT) {
      ignoreconstraints = "yes".equalsIgnoreCase(attributes.remove("ignoreconstraints"));
    }
    super.check();
    if ((name == null) && (type != BUTTON) && (type != RESET)) {
      attributes.requireValue("name");
    }
  }

  /*
  void scriptDefine(HtmlDocument document, PrintWriter out) {
    if (constraints != null) {
      getStatus();
      if (status != null) {
	status.scriptDefine(document, out);
      }
      getAutoComplete();
      if (autocomplete != null) {
	autocomplete.scriptDefine(document, out);
      }
    }
  }
  */

  /*
  void scriptInit(HtmlDocument document, PrintWriter out) {
    if ((constraints != null) && (status != null)) {
      status.scriptInit(document, out);
    }
    if (form != null) {
      String updateFunction = getDocument().getUpdateFunctionName(form.id, name);
      if ((updateFunction != null) || (initValue != null)) {
	out.println("  element = document.forms["+form.index+"].elements["+index+"];");
	if (id != null) {
	  out.println("  PowerForms.addElementById('"+id+"', element);");
	}
	if (updateFunction != null) {
	  if ((type == TEXT) || (type == PASSWORD)) {
	    out.println("  if (element.addEventListener) {");
	    out.println("    element.addEventListener('onkeyup', "+updateFunction+", true);");
	    out.println("    element.addEventListener('onchange', "+updateFunction+", true);");
	    out.println("  } else {");
	    out.println("    element.onkeyup = isAcademicBrowser ? function(e) { this.blur(); this.focus(); "+updateFunction+"; } : "+updateFunction+";");
	    out.println("    element.onchange = "+updateFunction+";");
	    out.println("  }");
	  } else if ((type == RADIO) || (type == CHECKBOX)) {
	    out.println("  if (element.addEventListener) {");
	    out.println("    element.addEventListener('onclick', "+updateFunction+", true);");
	    out.println("  } else {");
	    out.println("    element.onclick = "+updateFunction+";");
	    out.println("  }");
	    if (true || hasConstraints()) {
	      out.println("  initCheck(element);");
	    }
	  }
	}
	if (initValue != null) {
	  if (((type == TEXT) || (type == PASSWORD)) && (value == null)) {
	    out.println("  element.value = "+Html.scriptQuote(initValue)+";");
	  } else if (((type == RADIO) || (type == CHECKBOX))
		     && initValue.equals(value) && (attributes.getValueHtml("checked") == null)) {
	    out.println("  element.checked = true;");
	  }
	}
      }
      if ((constraints != null) && (autocomplete != null)) {
	autocomplete.scriptInit(document, out);
      }
    }
  }
  */

  void code(PrintWriter out) {
    out.print("<"+tagName);
    if (type == SUBMIT) {
      String submitFunctionName = (form == null ? null : form.getSubmitFunctionName());
      if (submitFunctionName != null) {
	out.print(" onclick='return "+submitFunctionName+"(");
	out.print(name == null ? "null" : "\""+name+"\"");
	/* out.print("null"); */
	out.print(", ");
	out.print(value == null ? "null" : "\""+value+"\"");
	/* out.print("null"); */
	out.print(", ");
	out.print(ignoreconstraints);
	out.print(");'");
      }
    } else if (type == RESET) {
      String resetFunctionName = (form == null ? null : form.getResetFunctionName());
      if (resetFunctionName != null) {
	out.print(" onclick='return "+resetFunctionName+"();'");
      }
    }
    attributes.code(out);
    out.print(">");
    content.code(out);
    out.print("</"+tagName+">");
  }
}


class HtmlSelect extends HtmlField {
  private String errorValue;
  ArrayList values;
  private ArrayList selected;

  HtmlSelect(Attributes attributes, HtmlContent content) {
    type = SELECT;
    init("select", attributes, content);
  }

  void check() throws TransformException {
    errorValue = attributes.remove("error");
    if (isNested()) {
      reportError("nested selects not allowed");
    }
    /* attributes.requireValue("name"); */
    values = new ArrayList();
    selected = new ArrayList();
    if (hasConstraints()) {
      getDocument().selectCode = true;
    }
    super.check();
    value = attributes.remove("value");
    if ((value == null) && (form != null)) {
      value = getPowerForms().getInitialization(form.id, name);
    }
  }

  void addValue(String value) {
    values.add(value);
  }

  void addSelected(String value) {
    selected.add(value);
  }

  void scriptInit(HtmlDocument document, PrintWriter out) {
    String updateFunction = (form == null ? null : getDocument().getUpdateFunctionName(form.id, name));
    if (updateFunction != null) {
      out.println("  element = document.forms["+form.index+"].elements["+index+"];");
      if (id != null) {
	out.println("  PowerForms.addElementById('"+id+"', element);");
      }
      String eventType = "change";
      out.println("  if (element.addEventListener) {");
      out.println("    element.addEventListener('"+eventType+"', "+updateFunction+", true);");
      out.println("  } else {");
      out.println("    element.on"+eventType+" = "+updateFunction+";");
      out.println("  }");
      if (true || hasConstraints()) {
	out.println("  initSelect(element);");
      }
    }
    super.scriptInit(document, out);
  }

  void code(PrintWriter out) {
    out.print("<"+tagName);
    attributes.removeHtml("value");
    attributes.code(out);
    out.print(">");
    content.code(out);
    out.print("</"+tagName+">");
  }
}


class HtmlTextArea extends HtmlField {
  HtmlTextArea(Attributes attributes, HtmlContent content) {
    type = TEXTAREA;
    init("textarea", attributes, content);
  }

  void check() throws TransformException {
    super.check();
    if (isNested()) {
      reportError("nested textareas not allowed");
    }
    attributes.requireValue("name");
  }

  void code(PrintWriter out) {
    out.print("<"+tagName);
    attributes.code(out);
    out.print(">");
    content.code(out);
    out.print("</"+tagName+">");
  }
}


class HtmlOption extends HtmlElement {
  private String value;
  private HtmlSelect select;
  private boolean selected = false;

  HtmlOption(Attributes attributes, HtmlContent content) {
    init("option", attributes, content);
  }

  void check() throws TransformException {
    select = (HtmlSelect)getAncestor(HtmlSelect.class);
    if (select != null) {
      value = attributes.getValueHtml("value");
      if (value == null) {
	StringBuffer v = new StringBuffer();
	for (int i = 0; i < content.size(); i++) {
	  v.append(((HtmlSomething)content.get(i)).getText());
	}
	value = v.toString().trim();
      }
      select.addValue(value);
      selected = attributes.hasValueHtml("selected");
      if (selected) {
	select.addSelected(value);
      }
    }
    super.check();
  }

  void script(HtmlDocument document, PrintWriter out) {}

  void code(PrintWriter out) {
    out.print("<"+tagName);
    attributes.code(out);
    if ((select != null) && value.equals(select.value) && !selected) {
      out.print(" selected='selected'");
    }
    out.print(">");
    content.print(out);
    out.print("</"+tagName+">");
  }
}


class HtmlApplet extends HtmlElement {
  private HtmlForm form;
  private int index;
  private ArrayList results;

  HtmlApplet(Attributes attributes, HtmlContent content) {
    init("applet", attributes, content);
  }

  void check() throws TransformException {
    index = getDocument().getAppletNumber();
    /* throw new LazyProgrammerException("Applet.check not yet implemented!"); */
  }

  void script(HtmlDocument document, PrintWriter out) {}

  void code(PrintWriter out) {
    /*
    for (int i = 0; i < results.size(); i++) {
      HtmlResult result = (HtmlResult)results.get(i);
      out.print("<input type='hidden' name='"+result.name+"' />");
    }
    */
    out.print("<"+tagName);
    attributes.code(out);
    out.print(">");
    content.print(out);
    out.print("</"+tagName+">");
  }

  void addResult(HtmlResult result) {
    results.add(result);
  }
}


class HtmlResult extends HtmlField {
  private HtmlApplet applet;

  HtmlResult(Attributes attributes) {
    init("result", attributes);
  }

  void check() throws TransformException {
    applet = (HtmlApplet)getAncestor(HtmlApplet.class);
    if (applet == null) {
      reportError("result outside applet");
    } else {
      attributes.requireValue("name");
      applet.addResult(this);
    }
  }

  void code(PrintWriter out) {}
}


class HtmlSubmit extends HtmlField {
  final static Set submitTypes = new HashSet();
  static {
    String[] types = { "anchor", "button", "submit" };
    for (int i = 0; i < types.length; i++) {
      submitTypes.add(types[i]);
    }
  }

  String submittype;
  boolean ignoreconstraints = false;

  HtmlSubmit(Attributes attributes, HtmlContent content) {
    type = SUBMIT_ELEMENT;
    init("submit", attributes, content);
  }

  void check() throws TransformException {
    super.check();
    submittype = attributes.getValue("type", "anchor");
    if (!submitTypes.contains(submittype)) {
      reportError("invalid submit type: `"+submittype+"'");
    }
    ignoreconstraints = "yes".equalsIgnoreCase(attributes.remove("ignoreconstraints"));
    if (form != null) {
      form.addSubmit(this);
    }
  }

  void script(HtmlDocument document, PrintWriter out) {}

  void code(PrintWriter out) {
    if (getDocument().generate_script_code) {
      if (submittype.equals("submit")) {
	out.print("<input type='submit'");
	if (name != null) {
	  out.print(" name='"+name+"'");
	}
	if (value != null) {
	  out.print(" value='"+value+"'");
	}
	String submitFunctionName = (form == null ? null : form.getSubmitFunctionName());
	if (submitFunctionName != null) {
	  out.print(" onclick='");
	  out.print(submitFunctionName+"(");
	  out.print(name == null ? "null" : "\""+name+"\"");
	  out.print(", ");
	  out.print(value == null ? "null" : "\""+value+"\"");
	  out.print(", ");
	  out.print(ignoreconstraints);
	  out.print(")'");
	}
	out.print(" />");
      } else {
	String submitFunctionName = (form == null ? null : form.getSubmitFunctionName());
	if (submittype.equals("button")) {
	  out.print("<input type='button' value=");
	  if ((content.size() == 0) && (value != null)) {
	    out.print(Html.scriptQuote(value));
	  } else {
	    out.print("'");
	    content.code(out);
	    out.print("'");
	  }
	  if ((submitFunctionName != null) || (form != null)) {
	    out.print(" onclick='");
	  }
	} else {
	  out.print("<a");
	  if ((submitFunctionName != null) || (form != null)) {
	    out.print(" onmouseover='window.status=\"\";' href='javascript:");
	  }
	}
	if (submitFunctionName != null) {
	  out.print(submitFunctionName+"(");
	  out.print(name == null ? "null" : "\""+name+"\"");
	  out.print(", ");
	  out.print(value == null ? "null" : "\""+value+"\"");
	  out.print(", ");
	  out.print(ignoreconstraints);
	  out.print(")");
	} else if (form != null) {
	  out.print("document.forms["+form.index+"].submit();");
	}
	if ((submitFunctionName != null) || (form != null)) {
	  out.print("'");
	}
	if (submittype.equals("button")) {
	  out.print(" />");
	} else {
	  out.print(">");
	  if ((content.size() == 0) && (value != null)) {
	    out.print(value);
	  } else {
	    content.code(out);
	  }
	  out.print("</a>");
	}
      }
    }
  }
}

class HtmlStatus extends HtmlElement {
  final static Set types = new HashSet();
  static {
    String[] v = { "image", "text", "css" };
    for (int i = 0; i < v.length; i++) {
      types.add(v[i]);
    }
  }

  protected String type;
  protected String error, warning, help;
  protected String red, yellow, green, na, blank;
  protected HtmlForm form = null;
  protected HtmlField field = null;
  protected Constraint constraint = null;
  private int index;

  HtmlStatus(Attributes attributes) {
    init("status", attributes);
  }

  void check() throws TransformException {
    attributes.addID();
    id = attributes.getValue("id");
    index = getDocument().getStatusNumber();
    type = attributes.getValue("type", "image");
    getDocument().statusCode = true;
    if (!types.contains(type)) {
      reportError("unknown status type: `"+type+"'");
    } else if (type.equals("image")) {
      URL miscUrl = getPowerForms().getMiscUrl();
      try {
	Attributes a = getPowerForms().getDefaultStatusAttributes();
	String red_default = a.getValue("red");
	String yellow_default = a.getValue("yellow");
	String green_default = a.getValue("green");
	String na_default = a.getValue("na");
	String blank_default = a.getValue("blank");
	red = new URL(miscUrl, attributes.getValue("red", red_default)).toString();
	yellow = new URL(miscUrl, attributes.getValue("yellow", yellow_default)).toString();
	green = new URL(miscUrl, attributes.getValue("green", green_default)).toString();
	na = new URL(miscUrl, attributes.getValue("na", na_default)).toString();
	blank = new URL(miscUrl, attributes.getValue("blank", blank_default)).toString();
      } catch (MalformedURLException e) {
	throw new TransformException(e);
      }
    } else if (type.equals("css")) {
      getDocument().cssCode = true;
      red = attributes.getValue("red", "pwfRed");
      yellow = attributes.getValue("yellow", "pwfYellow");
      green = attributes.getValue("green", "pwfGreen");
      na = attributes.getValue("na", "pwfNa");
    } else {
      red = attributes.getValue("red", "");
      yellow = attributes.getValue("yellow", "");
      green = attributes.getValue("green", "");
      na = attributes.getValue("na", "");
    }

    String idref = attributes.getValue("field");
    if (idref != null) {
      /* form = getDocument().getForm(attributes.getValue("form")); */
      HtmlElement element = getDocument().getElement(idref);
      /* if ((element == null) || !(element instanceof HtmlField) || (form != ((HtmlField)element).form)) { */
      if ((element == null) || !(element instanceof HtmlField)) {
	reportError("no field with id `"+idref+"'");
      } else {
	field = (HtmlField)element;
	form = field.form;
	field.setStatus(this);
      }
    } else {
      idref = attributes.getValue("constraint");
      if (idref != null) {
	constraint = getPowerForms().getConstraint(idref);
	if (constraint == null) {
	  reportError("no constraint with id `"+idref+"'");
	} else {
	  constraint.setStatus(this);
	}
      }
    }
    if (idref == null) {
      reportError("missing attribute `constraint' or 'field' for `"+tagName+"'");
    }
    super.check();
  }

  void setField(HtmlField field) {
    this.field = field;
  }

  String getScriptName() {
    return "pwf_status_"+index;
  }

  private boolean defined = false;
  private boolean initialized = false;

  void scriptDefine(HtmlDocument document, PrintWriter out) {
    if (!defined) {
      if ((constraint != null) || ((field != null) && field.hasConstraints())) {
	out.println("var "+getScriptName()+" = null;");
      }
      defined = true;
    }
  }

  void scriptInit(HtmlDocument document, PrintWriter out) {
    if (!initialized && defined) {
      out.println("  "+getScriptName()+" = new Status(PowerForms."+type.toUpperCase()+", "
		  +"'"+(type.equals("css") ? field.id : id)+"', "
		  +Html.scriptQuote(na)+", "
		  +Html.scriptQuote(red)+", "
		  +Html.scriptQuote(yellow)+", "
		  +Html.scriptQuote(green)+");");
      initialized = true;
    }
  }

  void code(PrintWriter out) {
    if (initialized) {
      if (type.equals("image")) {
	out.print("<img id='"+id+"' name='"+id+"' src='"+blank+"' alt='' />");
      } else if (type.equals("text")) {
	out.print("<span id='"+id+"' />");
      }
    }
  }
}


abstract class HtmlField extends HtmlElement {
  static final int BUTTON = 1;
  static final int CHECKBOX = 2;
  static final int FILE = 3;
  static final int HIDDEN = 4;
  static final int IMAGE = 5;
  static final int PASSWORD = 6;
  static final int RADIO = 7;
  static final int RESET = 8;
  static final int SUBMIT = 9;
  static final int TEXT = 10;
  static final int SELECT = 11;
  static final int TEXTAREA = 12;
  static final int RESULT = 13;
  static final int SUBMIT_ELEMENT = 14;

  int type;
  HtmlForm form;
  String name, value, initValue;
  int index;
  ListSet constraints = null;
  HtmlFieldList fieldList;

  void check() throws TransformException {
    name = attributes.getValueHtml("name");
    /*
    if ((name != null) && name.startsWith("#")) {
      name = name.substring(1);
    }
    */
    value = attributes.getValueHtml("value");
    form = getForm();
    if (form != null) {
      form.addField(this);
      constraints = getPowerForms().getConstraints(form.id, name);
      initValue = getPowerForms().getInitialization(form.id, name);
      if (((constraints != null) || (initValue != null)) && (id == null)) {
	if (attributes.addID()) {
	  getDocument().addElement(this);
	}
      }
    }
    super.check();
  }

  private HtmlForm getForm() {
    for (PwfElement p = getParent(); p != null; p = p.getParent()) {
      if (p instanceof HtmlForm) {
	return (HtmlForm)p;
      }
    }
    return null;
  }

  protected HtmlStatus status;

  HtmlStatus getStatus() {
    return status;
  }

  AutoComplete getAutoComplete() {
    return null;
  }

  void setStatus(HtmlStatus status) {
    this.status = status;
  }

  void setFieldList(HtmlFieldList fieldList) {
    this.fieldList = fieldList;
  }

  boolean hasConstraints() {
    /* return fieldList.constrainedBy.size() > 0; */
    return constraints != null;
  }

  /*
    void setConstraints(ArrayList constrainedBy, ArrayList affects) {
    this.constraints = constraints;
    }
  */

  void script(HtmlDocument document, PrintWriter out) {}

  public String toString() {
    /* return "["+tagName+" "+name+" ("+fieldList.constrainedBy+"; "+fieldList.affects+")]"; */
    return "["+tagName+" "+(name == null ? "no-name" : name)+"]";
  }
}


class HtmlFieldList {
  HtmlForm form;
  String name;
  ArrayList constrainedBy;
  ArrayList affects;
  ArrayList constraints;
  ArrayList fields;
  int lowerBound, upperBound;

  HtmlFieldList(HtmlForm form, String name) {
    this.form = form;
    this.name = name;
    constrainedBy = new ArrayList();
    affects = new ArrayList();
    constraints = new ArrayList();
    fields = new ArrayList();
  }

  void addField(HtmlField field) {
    if (field instanceof HtmlInput) {
      HtmlInput input = (HtmlInput)field;
      if (input.type == HtmlField.RADIO) {
	boolean firstRadio = true;
	for (int i = 0; i < size(); i++) {
	  HtmlField f = getField(i);
	  if (f.type == HtmlField.RADIO) {
	    firstRadio = false;
	    break;
	  }
	}
	if (firstRadio) {
	  // See ``http://www.w3.org/TR/html4/interact/forms.html#radio''
	  /* lowerBound++; */
	  upperBound++;
	}
      } else {
	upperBound++;
      }
      if ((input.type == HtmlField.TEXT) || (input.type == HtmlField.PASSWORD)
	  || (input.type == HtmlField.HIDDEN) || (input.type == HtmlField.TEXTAREA)) {
	lowerBound++;
      }
    } else if (field instanceof HtmlTextArea) {
      lowerBound++;
      upperBound++;
    } else if (field instanceof HtmlSelect) {
      HtmlSelect select = (HtmlSelect)field;
      if (select.attributes.getValueHtml("multiple") == null) {
	lowerBound++;
	upperBound++;
      } else {
	upperBound += select.values.size();
      }
    }
    fields.add(field);
    field.setFieldList(this);
  }

  HtmlField getField(int i) {
    return (HtmlField)fields.get(i);
  }

  int size() {
    return fields.size();
  }

  public String toString() {
    return "[HtmlFieldList: "+name+"; bounds: ["+lowerBound+", "+upperBound+"]; "+fields+"]";
  }
}

/* abstract class HtmlNode extends PwfNode {} */


class HtmlElement extends PwfElement {
  protected HtmlDocument document;

  protected HtmlElement() {}

  HtmlElement(String tagName, Attributes attributes, HtmlContent content) {
    init(tagName, attributes, content);
  }

  HtmlDocument getDocument() {
    return document;
  }

  PowerForms getPowerForms() {
    return getDocument().getPowerForms();
  }

  void setDocument(HtmlDocument document) {
    this.document = document;
    if (id != null) {
      document.addElement(this);
    }
    if (content != null) {
      ((HtmlContent)content).setDocument(document);
    }
  }

  private boolean isJWIGSubmitAnchor() {
    if (tagName.equalsIgnoreCase("a")) {
      String href = attributes.getValueHtml("href");
      return (href != null) && (href.startsWith("javascript:submitbutton("));
    }
    return false;
  }

  private HtmlForm jwigSubmitForm = null;
  private String jwigSubmitAnchorValue = null;
  private boolean jwigSubmitIgnoreConstraints = false;

  void check() throws TransformException {
    if (isJWIGSubmitAnchor()) {
      String href = attributes.getValueHtml("href");
      int start = href.indexOf('(');
      int firstComma = href.indexOf(',');
      int secondComma = href.indexOf(',', firstComma+1);
      int end = href.indexOf(')');
      String submitFormname = href.substring(start+1, firstComma).trim();
      submitFormname = submitFormname.substring(1, submitFormname.length()-1);
      jwigSubmitAnchorValue = href.substring(firstComma+1, secondComma > -1 ? secondComma : end).trim();
      jwigSubmitAnchorValue = jwigSubmitAnchorValue.substring(1, jwigSubmitAnchorValue.length()-1);
      jwigSubmitIgnoreConstraints = ((secondComma > -1) && "true".equalsIgnoreCase(href.substring(secondComma+1, end).trim()));
      String ignoreconstraints = attributes.remove("ignoreconstraints");
      if (!jwigSubmitIgnoreConstraints) {
	jwigSubmitIgnoreConstraints = (ignoreconstraints != null) && "yes".equalsIgnoreCase(ignoreconstraints.trim());
      }
      /* jwigSubmitAnchorValue = submitName.substring(1, submitName.length()-1); */
      /*
	System.err.println("JWIG submit anchor: "+this+"\t"+attributes);
	System.err.println("  Formname: "+submitFormname);
	System.err.println("  Name    : "+submitName);
	System.err.println();
      */
      jwigSubmitForm = getDocument().getForm(submitFormname);
      if (jwigSubmitForm != null) {
	jwigSubmitForm.addJWIGSubmitAnchor(jwigSubmitAnchorValue, jwigSubmitIgnoreConstraints);
      }
    }
    super.check();
  }

  void code(PrintWriter out) {
    if (jwigSubmitForm != null) {
      attributes.add("href", "javascript:void("+jwigSubmitForm.getSubmitFunctionName()+"('$submit', '"+jwigSubmitAnchorValue+"', "+jwigSubmitIgnoreConstraints+"))");
    }
    out.print("<"+tagName);
    if (attributes != null) {
      attributes.print(out);
    }
    if ((content == null) || (content.size() == 0)) {
      out.print("/>");
    } else {
      out.print(">");
      content.code(out);
      out.print("</"+tagName+">");
    }
  }

  boolean isNested() {
    return getAncestor(getClass()) != null;
  }
}


class HtmlContent extends PwfContent {
  void setDocument(HtmlDocument document) {
    for (int i = 0; i < size(); i++) {
      HtmlElement e = (HtmlElement)get(i);
      e.setDocument(document);
    }
  }
}
