/*
 * PowerForms
 * Copyright (C) 1999-2002 BRICS.
 *
 * Author: Mikkel Ricky <ricky@brics.dk>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the  Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 * USA.
 *
 * Home page: http://www.brics.dk/bigwig/powerforms/
 */

package dk.brics.powerforms;

import java.io.*;
import java.net.*;
import java.util.*;

%%

%class Lexer
%throws ParseException
%unicode
%line
%column
%int

// %debug
// %standalone

%{
  /* if ignore_whitespace_level > 0 then whitespace should be
     ignored. This holds when parsing formats, regexp, and constraints */
  int ignore_whitespace_level = 0;
  int token = YYEOF;

  int match(int kind) throws ParseException {
    if (token != kind) {
      throw new ParseException(this, "Expected: `"+Token.getImage(kind)+"'; found: `"+Token.getImage(token)+"'");
    }
    return nextToken();
  }

  void matchEndTag(int kind) throws ParseException {
    if (token == Token.ENDTAG) {
      switch (kind) {
      case Token.CONSTRAINT_end:
      /* case Token.FORMAT_end: */
      case Token.REGEXP_end:
	ignore_whitespace_level--;
	break;
      }
      match(Token.ENDTAG);
    } else {
      match('>');
      match(kind);
    }
  }

  int nextToken() throws ParseException {
    if (child != null) {
      token = child.nextToken();
      if (token == YYEOF) {
	// System.err.println("\tPopping "+next.url);
	child = null;
      } else {
	return token;
      }
    }
    try {
      return token = yylex();
    } catch (IOException e) {
      throw new ParseException(e.getMessage());
    }
  }

  boolean hasMoreTokens() {
    return token != YYEOF;
  }

  private URL url = null;

  URL getUrl() {
    if (child != null) {
      return child.getUrl();
    }
    return url;
  }

  int getLine() {
    if (child != null) {
      return child.getLine();
    }
    return yyline+1;
  }

  int getColumn() {
    if (child != null) {
      return child.getColumn();
    }
    return yycolumn;
  }

  String getLocation() {
    URL url = getUrl();
    if (url == null) {
      return "line "+getLine()+":"+getColumn();
    }
    return url+":"+getLine()+":"+getColumn();
  }

  private String text = null;

  String getText() {
    if (child != null) {
      return child.getText();
    }
    if (text != null) {
      String s = text;
      text = null;
      return s;
    }
    return yytext();
  }

  private void setText() {
    setText(yytext());
  }

  private void setText(String text) {
    this.text = text;
  }

  private boolean cyclicInclusion(URL url) {
    for (Lexer p = parent; p != null; p = p.parent) {
      if ((p.url != null) && url.sameFile(p.url)) {
	return true;
      }
    }
    return false;
  }

  private Lexer parent = null;
  private Lexer child = null;
  private String spec = null;

  Lexer(String spec) throws ParseException {
    this(spec, null);
  }

  private Lexer(String spec, Lexer parent) throws ParseException {
    try {
      Reader reader;
      if (spec.equals("-")) {
	reader = new InputStreamReader(System.in);
      } else {
	File file = null;
	try {
	  if ((parent == null) || (parent.url == null)) {
	    url = new URL(spec);
	  } else {
	    url = new URL(parent.url, spec);
	  }
	  file = File.createTempFile("powerforms", "Lexer");
	  file.createNewFile();
	  file.deleteOnExit();
	  FileWriter out = new FileWriter(file);
	  InputStream in = url.openStream();
	  int i;
	  while ((i = in.read()) != -1) {
	    out.write(i);
	  }
	  in.close();
	  out.close();
	} catch (MalformedURLException e) {
	  if ((parent == null) || (parent.url == null)) {
	    file = new File(spec);
	  } else {
	    File parentFile = new File(parent.url.getFile());
	    file = new File(parentFile.isFile() ? parentFile : null, spec);
	  }
	  try {
	    url = file.toURL();
	  } catch (MalformedURLException e1) {}
	}
	if (cyclicInclusion(url)) {
	  throw new RuntimeException("Cyclic inclusion of url ``"+url+"''!");
	}
	reader = new FileReader(file);
      }
      yyreset(reader);
    } catch (IOException e) {
      if (url == null) {
	throw new ParseException("Unable to "+((parent == null) || (parent.url == null) ? "open" : "include")+" "+spec);
      } else {
	throw new ParseException("Unable to "+((parent == null) || (parent.url == null) ? "open" : "include")+" url: "+url);
      }
    }
  }

  private boolean crapdebug = !true;
  private boolean crapmode = false;

  private int TAG = PWF_TAG;

  void setCrapMode(boolean crapmode) {
    this.crapmode = crapmode;
    TAG = crapmode ? CRAP_TAG : PWF_TAG;
  }

  private boolean expectAttributeValue = false;
%}

NameChar	= [[:letter:][:digit:]_:.-]
Name		= [[:letter:]_:]{NameChar}*
HtmlAttrValue	= [a-zA-Z0-9._:-]+

OCT		= [0-3]? [0-7]{1,2}
HEX		= (x|X)[0-9A-Fa-f]{1,2}
UNICODE		= u+[0-9A-Fa-f]{1,4}
ESCAPE		= \\(.|{OCT}|{HEX}|{UNICODE})

StringSingleQuote = \' ( [^\\\'] | {ESCAPE} )* \'
StringDoubleQuote = \" ( [^\\\"] | {ESCAPE} )* \"
String = {StringSingleQuote} | {StringDoubleQuote}

Comment = "<!--" ([^-] | "-"[^-])* "-"? "-->"

S        = [\u0020\u0009\u000D\u000A]+

Form     = [Ff][Oo][Rr][Mm]
Input    = [Ii][Nn][Pp][Uu][Tt]
Select   = [Ss][Ee][Ll][Ee][Cc][Tt]
TextArea = [Tt][Ee][Xx][Tt][Aa][Rr][Ee][Aa]
Option   = [Oo][Pp][Tt][Ii][Oo][Nn]
Applet   = [Aa][Pp][Pp][Ll][Ee][Tt]

%x		INCLUDE HTML_TAG PWF_TAG CRAP_TAG

%%

<YYINITIAL> {
  // Html that is interesting to PowerForms ...
  "<"  {Form}                   { yybegin(TAG); return Token.FORM_start; }
  "</" {Form} {S}? ">"		{               return Token.FORM_end; }
  "<"  {Input}         		{ yybegin(TAG); return Token.INPUT_start; }
  "</" {Input} {S}? ">" 	{               return Token.INPUT_end; }
  "<"  {Select}        		{ yybegin(TAG); return Token.SELECT_start; }
  "</" {Select} {S}? ">"	{               return Token.SELECT_end; }
  "<"  {TextArea}      		{ yybegin(TAG); return Token.TEXTAREA_start; }
  "</" {TextArea} {S}? ">"      {               return Token.TEXTAREA_end; }
  "<"  {Option}        		{ yybegin(TAG); return Token.OPTION_start; }
  "</" {Option}	{S}? ">" 	{               return Token.OPTION_end; }
  "<"  {Applet}        		{ yybegin(TAG); return Token.APPLET_start; }
  "</" {Applet} {S}? ">"	{               return Token.APPLET_end; }

  // The PowerForms tags ...
  /*
  "<result"       	{ yybegin(TAG); return Token.RESULT_start; }
  "</result>"		{               return Token.RESULT_end; }

  "<include"         	{ yybegin(INCLUDE); }
  "</include>"		{}

  "<constraint"	        { yybegin(TAG); ignore_whitespace_level++; return Token.CONSTRAINT_start; }
  "</constraint>"	{               ignore_whitespace_level--; return Token.CONSTRAINT_end; }
  "<constraint-def"	{ yybegin(TAG); ignore_whitespace_level++; return Token.CONSTRAINTDEF_start; }
  "</constraint-def>"	{               ignore_whitespace_level--; return Token.CONSTRAINTDEF_end; }

  "<powerforms>"       	{               ignore_whitespace_level++; return Token.POWERFORMS_start; }
  "</powerforms>"	{               ignore_whitespace_level--; return Token.POWERFORMS_end; }

  // "<format"         	{ yybegin(TAG); ignore_whitespace_level++; return Token.FORMAT_start; }
  // "</format>"        {               ignore_whitespace_level--; return Token.FORMAT_end; }

  "<if>"		{ return Token.IF_start; }
  "</if>"		{ return Token.IF_end; }
  "<then>"		{ return Token.THEN_start; }
  "</then>"		{ return Token.THEN_end; }
  "<else>"		{ return Token.ELSE_start; }
  "</else>"		{ return Token.ELSE_end; }
  "<ignore"		{ yybegin(TAG); return Token.IGNORE_start; }
  "</ignore>"		{               return Token.IGNORE_end; }

  // Expressions
  "<count"  	        { yybegin(TAG); return Token.COUNT_start; }
  "</count>"		{               return Token.COUNT_end; }
  "<match"  	        { yybegin(TAG); return Token.MATCH_start; }
  "</match>"		{               return Token.MATCH_end; }
  "<matchall"       	{ yybegin(TAG); return Token.MATCHALL_start; }
  "</matchall>"		{               return Token.MATCHALL_end; }
  "<equal"	        { yybegin(TAG); return Token.EQUAL_start; }
  "</equal>"		{               return Token.EQUAL_end; }
  "<function"	        { yybegin(TAG); return Token.FUNCTION_start; }
  "</function>"		{               return Token.FUNCTION_end; }
  "<and>"		{ return Token.AND_start; }
  "</and>"		{ return Token.AND_end; }
  "<or>"		{ return Token.OR_start; }
  "</or>"		{ return Token.OR_end; }
  "<not>"		{ return Token.NOT_start; }
  "</not>"		{ return Token.NOT_end; }

  // Regular expressions
  "<const"	        { yybegin(TAG); return Token.CONST_start; }
  "</const>"		{ 	        return Token.CONST_end; }
  "<empty"		{ yybegin(TAG); return Token.EMPTY_start; }
  "</empty>"		{ 	        return Token.EMPTY_end; }
  "<anychar"		{ yybegin(TAG); return Token.ANYCHAR_start; }
  "</anychar>"		{      	        return Token.ANYCHAR_end; }
  "<anything"		{ yybegin(TAG); return Token.ANYTHING_start; }
  "</anything>"		{ 	        return Token.ANYTHING_end; }
  "<interval"		{ yybegin(TAG); return Token.INTERVAL_start; }
  "</interval>"		{ 	        return Token.INTERVAL_end; }
  "<charrange"		{ yybegin(TAG); return Token.CHARRANGE_start; }
  "</charrange>"	{ 	        return Token.CHARRANGE_end; }
  "<range"		{ yybegin(TAG); return Token.CHARRANGE_start; }
  "</range>"		{ 	        return Token.CHARRANGE_end; }
  "<intersection>"	{ return Token.INTERSECTION_start; }
  "</intersection>"	{ return Token.INTERSECTION_end; }
  "<concat>"		{ return Token.CONCAT_start; }
  "</concat>"		{ return Token.CONCAT_end; }
  "<union>"		{ return Token.UNION_start; }
  "</union>"		{ return Token.UNION_end; }
  "<repeat"		{ yybegin(TAG); ignore_whitespace_level++; return Token.REPEAT_start; }
  "</repeat>"		{               ignore_whitespace_level--; return Token.REPEAT_end; }
  // "<star>"		{ return Token.STAR_start; }
  // "</star>"		{ return Token.STAR_end; }
  "<plus>"		{ return Token.PLUS_start; }
  "</plus>"		{ return Token.PLUS_end; }
  "<optional>"		{ return Token.OPTIONAL_start; }
  "</optional>"		{ return Token.OPTIONAL_end; }
  "<complement>"	{ return Token.COMPLEMENT_start; }
  "</complement>"	{ return Token.COMPLEMENT_end; }
  "<charset"		{ yybegin(TAG); return Token.CHARSET_start; }
  "</charset>"		{ 	        return Token.CHARSET_end; }
  "<regexp"		{ yybegin(TAG); ignore_whitespace_level++; return Token.REGEXP_start; }
  "</regexp>"		{               ignore_whitespace_level--; return Token.REGEXP_end; }
  "<regexp-def"		{ yybegin(TAG); ignore_whitespace_level++; return Token.REGEXPDEF_start; }
  "</regexp-def>"	{               ignore_whitespace_level--; return Token.REGEXPDEF_end; }

  "<init"               { yybegin(TAG); return Token.INIT_start; }
  "</init>"		{               return Token.INIT_end; }
  */

  "<submit"             { yybegin(TAG); return Token.SUBMIT_start; }
  "</submit>"		{               return Token.SUBMIT_end; }
  "<status"             { yybegin(TAG); return Token.STATUS_start; }
  "</status>"		{               return Token.STATUS_end; }

  {Comment}             { if (ignore_whitespace_level == 0) {
			    setText();
                            return Token.SOMETHING;
                          }
                        }

  "<" {Name}		{ yybegin(HTML_TAG); setText(); return Token.SOMETHING; }
  "</" {Name} ">"       {                    setText(); return Token.SOMETHING; }
  "<" | [^<]+		{ if (ignore_whitespace_level == 0) {
			    setText();
			    return Token.SOMETHING;
			  }
			}
}

<HTML_TAG> {
  [^<>]*		{ setText();
  			  return Token.SOMETHING;
  			}
  [<>]			{ yybegin(YYINITIAL);
  			  setText();
			  return Token.SOMETHING;
  			}
}

<PWF_TAG> {
  "="			{ return '='; }
  {Name}		{ return Token.Name; }
  {String}              { setText(yytext().substring(1, yylength()-1));
  			  return Token.AttValue;
			}
  {HtmlAttrValue}	{ return Token.AttValue; }

  "/>"			{ yybegin(YYINITIAL);
			  return Token.ENDTAG;
			}
  ">"			{ yybegin(YYINITIAL);
			  return '>';
			}
  [ \t\r\n\f]+		{ }
}

<CRAP_TAG> {
  "="			{ expectAttributeValue = true;
                          return '=';
                        }
  {Name}		{ if (crapmode && expectAttributeValue) {
			    expectAttributeValue = false;
			    if (crapdebug) { System.err.println(getLine()+": name as value: '"+yytext()+"'"); }
			    return Token.AttValue;
                          } else {
			    return Token.Name;
                          }
                        }
  {String}              { setText(yytext().substring(1, yylength()-1));
  			  expectAttributeValue = false;
  			  return Token.AttValue;
			}

  // A feeble attempt to parse the kind of html some (most?) people write ...
  ( [^ \t\r\n\f/>] | "/" [^>] )* {
    if (crapdebug) { System.err.println(getLine()+": HtmlAttrValue: '"+yytext()+"'"); }
    int eqIndex = yytext().indexOf('=');
    if (!expectAttributeValue && (eqIndex > -1)) {
      if (eqIndex == 0) {
	if (crapdebug) { System.err.println(getLine()+":   Pushing back '"+yytext().substring(1)+"' ..."); }
	yypushback(yylength()-1);
	if (crapdebug) { System.err.println(getLine()+":     equal: '"+yytext()+"'"); }
	expectAttributeValue = true;
	return '=';
      } else {
	if (crapdebug) { System.err.println(getLine()+":   Pushing back '"+yytext().substring(eqIndex)+"' ..."); }
	yypushback(yylength()-eqIndex);
	if (crapdebug) { System.err.println(getLine()+":     name: '"+yytext()+"'"); }
	return Token.Name;
      }
    } else {
      if (crapdebug) { System.err.println(getLine()+":     value: '"+yytext()+"'"); }
      if (Character.isWhitespace(yycharat(yylength()-1))) {
	yypushback(1);
      }
      expectAttributeValue = false;
      return Token.AttValue;
    }
  }

  "/>"			{ yybegin(YYINITIAL);
			  expectAttributeValue = false;
			  return Token.ENDTAG;
			}
  ">"			{ yybegin(YYINITIAL);
  			  expectAttributeValue = false;
			  return '>';
			}
  [ \t\r\n\f]+		{ expectAttributeValue = false; }
}

<INCLUDE> {
  "href" {S}? "=" {S}? {String} {
    int index = yytext().indexOf('=');
    if (index < 0) {
      // System.err.println("JFlex bug: (\"href\" {S}? \"=\" {S}? {String}) matches ("+yytext()+")!");
    } else {
      String s = yytext().substring(yytext().indexOf('=')+1).trim();
      String spec = s.substring(1, s.length()-1).trim();
      child = new Lexer(spec, this);
    }
  }
  "/"? ">" { yybegin(YYINITIAL); }
  [^]                          {}
}

/*
[^] {
  throw new ParseException("Unmatched input at "+getLocation()+": "+yytext());
}
*/

/*
static String normalize(String s) {
  char[] c = s.toCharArray();
  StringBuffer sb = new StringBuffer();
  for (int i = 0; i < c.length; i++) {
    if (Character.isWhitespace(c[i])) {
      if ((i < c.length-1) && Character.isWhitespace(c[i+1])) {
        while (Character.isWhitespace(c[i+1])) {
	  i++;
	}
      }
      sb.append(' ');
    } else {
      sb.append(c[i]);
    }
  }
  // System.err.println("normalize("+s+") = "+sb);
  return sb.toString();
}

static String decode(String s) {
  char[] c = s.toCharArray();
  StringBuffer sb = new StringBuffer();
  for (int i = 0; i < c.length; i++) {
    switch (c[i]) {
    case '\\':
	    switch (c[i+1]) {
	    case 'b':
		sb.append('\b');
		break;
	    case 't':
		sb.append('\t');
		break;
	    case 'n':
		sb.append('\n');
		break;
	    case 'f':
		sb.append('\f');
		break;
	    case 'r':
		sb.append('\r');
		break;
	    case '\'':
		sb.append('\'');
		break;
	    case '"':
		sb.append('"');
		break;
	    case '\\':
		sb.append('\\');
		break;
	    case 'u':
		sb.append((char)Integer.parseInt(s.substring(i+2, i+6), 16));
		i += 4;
		break;
	    default:
		if ((('0' <= c[i+1]) && (c[i+1] <= '3')) &&
		    (('0' <= c[i+2]) && (c[i+2] <= '7')) &&
		    (('0' <= c[i+3]) && (c[i+3] <= '7'))) {
		    sb.append((char)(64*(c[i+1]-'0')+8*(c[i+2]-'0')+(c[i+3]-'0')));
		    i += 2;
		} else if ((('0' <= c[i+1]) && (c[i+1] <= '7')) &&
			   (('0' <= c[i+2]) && (c[i+2] <= '7'))) {
		    sb.append((char)(8*(c[i+1]-'0')+(c[i+2]-'0')));
		    i++;
		} else if (('0' <= c[i+1]) && (c[i+1] <= '7')) {
		    sb.append((char)(c[i+1]-'0'));
		}
		break;
	    }
	    i++;
	    break;
	default:
	    sb.append(c[i]);
	    break;
	}
  }
  // System.err.println("decode("+s+") = "+sb);
  return sb.toString();
}

*/
