/*
 * PowerForms
 * Copyright (C) 1999-2002 BRICS.
 *
 * Author: Mikkel Ricky <ricky@brics.dk>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the  Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 * USA.
 *
 * Home page: http://www.brics.dk/bigwig/powerforms/
 */

package dk.brics.powerforms;

import dk.brics.automaton.*;
import java.io.*;
import java.util.*;

class Main {
  private static String[] inputs = null;
  private static String dotname = null;;
  private static String exportname = null;;
  private static String outputfilename = null;
  private static String errorfilename = null;
  private static String pwffilename = null;
  private static boolean verbose = false;
  private static boolean dot = false;

  static void usage(boolean error) {
    PrintStream out = error ? System.err : System.out;
    out.println("Usage: "+Main.class.getName()+" [options] [inputfile]");
    out.println();
    out.println("Options:");
    out.println("  -pwf <file>             specify powerforms file");
    out.println("  -export <name>          export regular expression named <name> to stdout");
    out.println("  -help                   print this message and exit");
    /* out.println("  -info                   print info about progress on the standard output stream"); */
    out.println("  -miscurl <url>          set url from which to include images and generic script code");
    out.println("                            (default: `"+Config.MISCURL+"')");
    out.println("  -status (image|css)     default status type");
    out.println("  -output <file>          write output to <file>");
    /* out.println("  -xhtml                  parse html as xhtml"); */
    /* out.println("  -no-output              don't generate html code"); */
    /* out.println("  -no-script              don't generate any script code"); */
    /* out.println("  -no-logo                don't insert PowerForms logo"); */
    out.println("  -version                print version information and exit");
    out.println("");
    out.println("Undocumented options:");
    out.println("  -crap-mode              try to parse anything as html ...");
    out.println("");
    out.println("Input is read from the standard input stream if no inputfile is given ");
    out.println("or if `-' is given as an inputfile.");
    System.exit(error ? 1 : 0);
  }

  static String[] setOptions(String[] args, PowerForms pwf) {
    int index = 0;
    ArrayList inputs = new ArrayList();

    while (index < args.length) {
      boolean ok = true;
      String arg = args[index++].intern();
      try {
	if (arg == "-errorfile") {
	  errorfilename = args[index++];
	} else if (arg == "-dot") {
	  dotname = args[index++];
	} else if (arg == "-export") {
	  exportname = args[index++];
	} else if (arg == "-help") {
	  usage(false);
	} else if (arg == "-xhtml") {
	  pwf.setOption(PowerForms.XHTML_OPTION, true);
	} else if (arg == "-miscurl") {
	  pwf.setOption(PowerForms.MISCURL_OPTION, args[index++]);
	} else if (arg == "-status") {
	  String type = args[index++];
	  Attributes attributes = new Attributes(pwf);
	  attributes.add("type", type);
	} else if (arg == "-output") {
	  outputfilename = args[index++];
	} else if (arg == "-no-output") {
	  pwf.setOption(PowerForms.NO_OUTPUT_OPTION, true);
	} else if (arg == "-no-script") {
	  pwf.setOption(PowerForms.NO_SCRIPT_OPTION, true);
	} else if (arg == "-no-logo") {
	  pwf.setOption(PowerForms.NO_LOGO_OPTION, true);
	} else if (arg == "-version") {
	  System.out.println(Config.NAME+" "+Config.VERSION);
	  System.exit(0);
	} else if (arg == "-verbose") {
	  verbose = true;
	} else if (arg == "-dot") {
	  dot = true;
	} else if (arg == "-crap-mode") {
	  pwf.setOption(PowerForms.CRAP_MODE_OPTION, true);
	} else if (arg == "-pwf") {
	  pwffilename = args[index++];
	} else if (arg.startsWith("-") && !arg.equals("-")) {
	  System.err.println("unknown option: `"+arg+"'");
	  System.exit(1);
	} else {
	  inputs.add(arg);
	}
      } catch (ArrayIndexOutOfBoundsException e) {
	System.err.println("Missing argument for option `"+arg+"'");
	System.exit(1);
      } catch (Exception e) {
	System.err.println(e);
      }
    }
    return (String[])inputs.toArray(new String[inputs.size()]);
  }

  public static void main(String[] args) {
    PowerForms pwf = new PowerForms();
    inputs = setOptions(args, pwf);
    try {
      Reader in = null;
      if (pwffilename != null) {
	try {
	  in = new FileReader(pwffilename);
	  pwf.initialize(in);
	  in.close();
	} catch (IOException e) {
	  System.err.println("cannot open file `"+pwffilename+"' for reading");
	  System.exit(1);
	}
      }

      if (verbose) {
	System.err.println("powerforms: "+pwf);
      }
      if (dot) {
	System.err.println(pwf.toDot());
      }

      if (exportname != null) {
	Automaton a = pwf.getAutomaton(exportname);
	if (a == null) {
	  System.err.println("No regexp named `"+exportname+"'");
	  System.exit(1);
	}
	try {
	  a.store(System.out);
	  System.out.flush();
	  System.exit(0);
	} catch (IOException e) {}
      }

      if (dotname != null) {
	Automaton a = pwf.getAutomaton(dotname);
	if (a == null) {
	  System.err.println("No regexp named `"+dotname+"'");
	  System.exit(1);
	}
	System.out.println(a.toDot());
	System.out.flush();
	System.exit(0);
      }

      PrintWriter out = new PrintWriter(new OutputStreamWriter(System.out), true);
      if (outputfilename != null) {
	try {
	  out = new PrintWriter(new FileWriter(outputfilename), true);
	} catch (IOException e) {
	  System.err.println("cannot open file `"+outputfilename+"' for writing");
	  System.exit(1);
	}
      }

      if ((inputs.length == 0) || (inputs[0].equals("-"))) {
	in = new InputStreamReader(System.in);
      } else {
	try {
	  in = new FileReader(inputs[0]);
	} catch (IOException e) {
	  System.err.println("cannot open file `"+inputs[0]+"' for reading");
	  System.exit(1);
	}
      }
      HtmlDocument document = pwf.transform(in, out);

      if (verbose) {
	System.err.println("document: "+document);
      }
    } catch (Exception e) {
      System.err.println(e);
      /* e.printStackTrace(System.err); */
      System.exit(1);
    }
  }
}
