/*
 * PowerForms
 * Copyright (C) 1999-2002 BRICS.
 *
 * Author: Mikkel Ricky <ricky@brics.dk>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the  Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 * USA.
 *
 * Home page: http://www.brics.dk/bigwig/powerforms/
 */

package dk.brics.powerforms;

import java.net.*;
import java.util.*;
import org.jdom.*;

class Parser {
  private static void elementExpected(String name) throws ParseException {
    throw new ParseException("`"+name+"' element expected");
  }

  private static void missingAttribute(String attributeName, String elementName) throws ParseException {
    throw new ParseException("missing attribute `"+attributeName+"' for `"+elementName+"' element");
  }

  private static void nontermExpected(String name) throws ParseException {
    throw new ParseException(name+" expected");
  }

  /* Parsing from JDom tree */
  static void parsePowerForms(Document document, PowerForms pwf) throws ParseException, TransformException {
    Element e = document.getRootElement();
    String name = e.getName().intern();
    if (name != "powerforms") {
      throw new ParseException("root element must be powerforms");
    }
    for (Iterator i = e.getChildren().iterator(); i.hasNext(); ) {
      Element child = (Element)i.next();
      name = child.getName().intern();
      if (name == "regexp") {
	RegExpDef regexp = new RegExpDef(parseAttributes(child.getAttributes(), pwf), parseSubRegExp(child, pwf));
	pwf.addRegExpDef(regexp);
	/* pwf.addRegExpDef((RegExpDef)parseRegExp(child, pwf)); */
	/* } else if (name == "constraint-def") { */
	/* pwf.addConstraintDef((ConstraintDef)parseConstraint(child, pwf)); */
      } else if (name == "constraint") {
	Constraint c = (child.getChildren().size() != 1 ? null : parseConstraint((Element)child.getChildren().get(0), pwf));
	if (c == null) {
	  nontermExpected("constraint");
	}
	pwf.addConstraint(new ConstraintApply(parseAttributes(child.getAttributes(), pwf), c));
	/* pwf.addConstraint((ConstraintApply)parseConstraint(child, pwf)); */
      } else if (name == "init") {
	pwf.addInitialization(parseInit(child, pwf));
      } else if (name == "status") {
	pwf.setDefaultStatusAttributes(parseAttributes(child.getAttributes(), pwf));
      } else if (name == "autocomplete") {
	pwf.addAutoComplete(new AutoComplete(parseAttributes(child.getAttributes(), pwf)));
      } else if (name == "option") {
	String optionName = child.getAttributeValue("name");
	if (optionName == null) {
	  missingAttribute("name", name);
	}
	String optionValue = child.getAttributeValue("value");
	if (optionValue == null) {
	  missingAttribute("value", name);
	}
	if (optionName.equals("miscurl")) {
	  pwf.setOption(PowerForms.MISCURL_OPTION, optionValue);
	} else if (optionName.equals("errortitle")) {
	  pwf.setOption(PowerForms.ERRORTITLE_OPTION, optionValue);
	} else {
	  throw new ParseException("invalid name `"+optionName+"' for "+name+" element");
	}
      } else {
	throw new ParseException("`regexp', `constraint', `status', `init', `option', or `autocomplete' element expected");
      }
    }
  }

  static HtmlDocument parseHtmlDocument(Document d, PowerForms pwf) throws ParseException {
    HtmlDocument document = new HtmlDocument(pwf);
    HtmlContent content = parseHtmlContent(d.getContent(), pwf);
    DocType docType = d.getDocType();
    if (docType != null) {
      String s = docType.toString();
      document.setDocType(s.substring(s.indexOf('<'), s.lastIndexOf('>')+1));
    }
    document.setContent(content);
    return document;
  }

  private static String getName(Attribute a) {
    String namespace = a.getNamespacePrefix();
    return (namespace.length() > 0 ? namespace+":" : "")+a.getName();
  }

  private static Attributes parseAttributes(List l, PowerForms pwf) throws ParseException {
    Attributes attributes = new Attributes(pwf);
    for (Iterator i = l.iterator(); i.hasNext(); ) {
      Attribute attribute = (Attribute)i.next();
      attributes.add(getName(attribute), attribute.getValue());
    }
    return attributes;
  }

  private static Constraint parseConstraint(Element e, PowerForms pwf) throws ParseException {
    Constraint c = null;
    String name = e.getName().intern();
    Attributes attributes = parseAttributes(e.getAttributes(), pwf);

    /*
      if (name == "constraint") {
      if (e.getParent().isRootElement()) {
      Constraint constraint = e.hasChildren() ? parseConstraint((Element)e.getChildren().get(0), pwf) : null;
      if (constraint == null) {
      nontermExpected("constraint");
      }
      c = new ConstraintApply(attributes, constraint);
      } else {
      c = new ConstraintRef(attributes);
      }
      } else if (name == "constraint-def") {
      c = e.hasChildren() ? new ConstraintDef(attributes, parseConstraint((Element)e.getChildren().get(0), pwf)) : null;
      } else
    */
    if (name == "if") {
      List children = e.getChildren();
      Expression test = null;
      if (children.size() > 0) {
	test = parseExpression((Element)children.get(0), pwf);
      }
      if (test == null) {
	nontermExpected("expression");
      }
      Constraint thenpart = null;
      if ((children.size() > 1) && (children.get(1) instanceof Element)) {
	Element t = (Element)children.get(1);
	if (t.getName().equals("then")) {
	  if ((t.getChildren().size() == 1)
	      && (t.getChildren().get(0) instanceof Element)) {
	    t = (Element)t.getChildren().get(0);
	    thenpart = parseConstraint(t, pwf);
	    if (thenpart == null) {
	      RegExp r = parseRegExp(t, pwf);
	      if (r != null) {
		thenpart = new ConstraintExpression(new ExpressionMatch(new Attributes(pwf), r));
	      }
	    }
	  }
	  if (thenpart == null) {
	    nontermExpected("constraint");
	  }
	} else {
	  elementExpected("then");
	}
      } else {
	elementExpected("then");
      }
      Constraint elsepart = null;
      if ((children.size() > 2) && (children.get(2) instanceof Element)) {
	Element t = (Element)children.get(2);
	if (t.getName().equals("else")) {
	  if ((t.getChildren().size() == 1)
	      && (t.getChildren().get(0) instanceof Element)) {
	    t = (Element)t.getChildren().get(0);
	    elsepart = parseConstraint(t, pwf);
	    if (elsepart == null) {
	      RegExp r = parseRegExp(t, pwf);
	      if (r != null) {
		elsepart = new ConstraintExpression(new ExpressionMatch(new Attributes(pwf), r));
	      }
	    }
	  }
	  if (elsepart == null) {
	    nontermExpected("constraint");
	  }
	} else {
	  elementExpected("else");
	}
      } else {
	elementExpected("else");
      }
      c = new ConstraintIf(test, thenpart, elsepart);
    } else if (name == "ignore") {
      c = new ConstraintIgnore(attributes);
    } else {
      Expression expression = parseExpression(e, pwf);
      if (expression != null) {
	c = new ConstraintExpression(expression);
      } else {
	RegExp regexp = parseRegExp(e, pwf);
	if (regexp != null) {
	  c = new ConstraintExpression(new ExpressionMatch(new Attributes(pwf), regexp));
	}
      }
    }
    return c;
  }

  private static Expression parseExpression(Element e, PowerForms pwf) throws ParseException {
    Expression expression = null;
    String name = e.getName().intern();
    Attributes attributes = parseAttributes(e.getAttributes(), pwf);

    if (name == "count") {
      expression = new ExpressionCount(attributes);
    } else if (name == "match") {
      RegExp r = e.getChildren().size() != 1 ? null : parseRegExp((Element)e.getChildren().get(0), pwf);
      if (r == null) {
	nontermExpected("regexp");
      }
      expression = new ExpressionMatch(attributes, r);
      /* } else if (name == "match-all") { */
      /* expression = new ExpressionMatchAll(attributes, parseRegExp((Element)e.getChildren().get(0))); */
    } else if (name == "equal") {
      expression = new ExpressionEqual(attributes, parseFieldList(e.getChildren(), pwf));
    } else if (name == "less-than") {
      expression = new ExpressionLessThan(attributes, parseFieldList(e.getChildren(), pwf));
    } else if (name == "function") {
      expression = new ExpressionFunction(attributes);
    } else if (name == "and") {
      expression = new ExpressionAnd(parseExpressionContent(e.getChildren(), pwf));
    } else if (name == "or") {
      expression = new ExpressionOr(parseExpressionContent(e.getChildren(), pwf));
    } else if (name == "not") {
      expression = new ExpressionNot(parseExpressionContent(e.getChildren(), pwf));
    }
    return expression;
  }

  private static PwfContent parseFieldList(List l, PowerForms pwf) throws ParseException {
    PwfContent content = new PwfContent();
    for (Iterator i = l.iterator(); i.hasNext(); ) {
      Element element = (Element)i.next();
      ExpressionField field = parseField(element, pwf);
      if (field != null) {
	content.add(field);
      } else {
	elementExpected("field");
      }
    }
    return content;
  }

  private static ExpressionField parseField(Element e, PowerForms pwf) throws ParseException {
    if (e.getName().equals("field")) {
      return new ExpressionField(parseAttributes(e.getAttributes(), pwf));
    }
    return null;
  }

  private static PwfContent parseExpressionContent(List l, PowerForms pwf) throws ParseException {
    PwfContent content = new PwfContent();
    for (Iterator i = l.iterator(); i.hasNext(); ) {
      Element element = (Element)i.next();
      Expression expression = parseExpression(element, pwf);
      if (expression != null) {
	content.add(expression);
      } else {
	nontermExpected("expression");
      }
    }
    return content;
  }

  private static String getName(Element e) {
    String namespace = e.getNamespacePrefix();
    return (namespace.length() > 0 ? namespace+":" : "")+e.getName();
  }

  private static PwfNode parseHtml(Object o, PowerForms pwf) throws ParseException {
    PwfNode h = null;

    if (o instanceof Comment) {
      Comment comment = (Comment)o;
      h = new HtmlSomething("<!--"+comment.getText()+"-->");
    } else if (o instanceof ProcessingInstruction) {
      ProcessingInstruction pi = (ProcessingInstruction)o;
      h = new HtmlSomething("<?"+pi.getTarget()+" "+pi.getData()+"?>");
    } else if (o instanceof CDATA) {
      CDATA cdata = (CDATA)o;
      h = new HtmlSomething(cdata.getText());
    } else if (o instanceof EntityRef) {
      EntityRef ref = (EntityRef)o;
      h = new HtmlSomething(ref.getName());
    } else if (o instanceof Text) {
      Text text = (Text)o;
      h = new HtmlSomething(text.getText());
    } else if (o instanceof String) {
      h = new HtmlSomething(o.toString());
    } else if (o instanceof Element) {
      Element element = (Element)o;
      String name = getName(element).toLowerCase().intern();
      Attributes attributes = parseAttributes(element.getAttributes(), pwf);
      List namespaces = element.getAdditionalNamespaces();
      for (Iterator i = namespaces.iterator(); i.hasNext(); ) {
	Namespace ns = (Namespace)i.next();
	if (!"".equals(ns.getPrefix())) {
	  attributes.add("xmlns:"+ns.getPrefix(), ns.getURI());
	}
      }
      boolean addID = !true;
      HtmlContent content = new HtmlContent();
      if (name == "form") {
	content = parseHtmlContent(element.getContent(), pwf);
	h = new HtmlForm(attributes, content);
      } else if (name == "input") {
	h = new HtmlInput(attributes);
      } else if (name == "button") {
	content = parseHtmlContent(element.getContent(), pwf);
	h = new HtmlButton(attributes, content);
      } else if (name == "select") {
	content = parseHtmlContent(element.getChildren(), pwf);
	h = new HtmlSelect(attributes, content);
      } else if (name == "textarea") {
	content = parseHtmlContent(element.getContent(), pwf);
	h = new HtmlTextArea(attributes, content);
      } else if (name == "option") {
	content = parseHtmlContent(element.getContent(), pwf);
	h = new HtmlOption(attributes, content);
      } else if (name == "applet") {
	content = parseHtmlContent(element.getContent(), pwf);
	h = new HtmlApplet(attributes, content);
      } else if (name == "result") {
	h = new HtmlResult(attributes);
      } else if (name == "status") {
	h = new HtmlStatus(attributes);
      } else if (name == "submit") {
	content = parseHtmlContent(element.getContent(), pwf);
	h = new HtmlSubmit(attributes, content);
      } else {
	h = new HtmlElement(getName(element), attributes, parseHtmlContent(element.getContent(), pwf));
	addID = false;
      }
      if ((attributes != null) && addID) {
	attributes.addID();
      }
    }
    return h;
  }

  private static HtmlContent parseHtmlContent(List l, PowerForms pwf) throws ParseException {
    HtmlContent content = new HtmlContent();
    HtmlSomething something = null;

    for (Iterator i = l.iterator(); i.hasNext(); ) {
      Object node = i.next();
      PwfNode html = parseHtml(node, pwf);
      if (html != null) {
	content.add(html);
      } else {
	nontermExpected("html");
      }
    }
    /*
      while (firstHtml(lexer.token)) {
      if (lexer.token == Token.POWERFORMS_start) {
      lexer.match(Token.POWERFORMS_start);
      Content c = parseHtmlContent(lexer);
      for (int i = 0; i < c.size(); i++) {
      content.add(c.get(i));
      }
      lexer.match(Token.POWERFORMS_end);
      } else {
      Html h = parseHtml(lexer);
      if (h instanceof HtmlSomething) {
      if (something != null) {
      something.append((HtmlSomething)h);
      } else {
      content.add(h);
      something = (HtmlSomething)h;
      }
      } else {
      content.add(h);
      something = null;
      }
      }
      }
      // Remove whitespace between PowerForms elements ...
      if (true) {
      for (int i = 2; i < content.size(); ) {
      Html c0 = content.get(i-2);
      Html c1 = content.get(i-1);
      Html c2 = content.get(i);
      if ((c0.getClass() == Html.class)
      && (c1 instanceof HtmlSomething)
      && (c2.getClass() == Html.class)
      && (((HtmlSomething)c1).getText().trim().length() == 0)) {
      content.remove(i-1);
      } else {
      i++;
      }
      }
      }
    */
    return content;
  }

  private static Init parseInit(Element e, PowerForms pwf) throws ParseException {
    Init init = null;
    String name = e.getName().intern();
    Attributes attributes = parseAttributes(e.getAttributes(), pwf);

    if (name == "init") {
      init = new Init(attributes);
    }
    return init;
  }

  private static RegExp parseSubRegExp(Element e, PowerForms pwf) throws ParseException {
    if (e.hasChildren() && (e.getChildren().get(0) instanceof Element)) {
      RegExp r = parseRegExp((Element)e.getChildren().get(0), pwf);
      if (r != null) {
	return r;
      }
    }
    nontermExpected("regexp");
    return null;
  }

  private static RegExp parseRegExp(Element e, PowerForms pwf) throws ParseException {
    RegExp regexp = null;
    String name = e.getName().intern();
    Attributes attributes = parseAttributes(e.getAttributes(), pwf);
    if (name == "const") {
      regexp = new RegExpConst(attributes);
    } else if (name == "empty") {
      regexp = new RegExpEmpty();
    } else if (name == "anychar") {
      regexp = new RegExpAnyChar();
    } else if (name == "anything") {
      regexp = new RegExpAnything();
    } else if (name == "charset") {
      regexp = new RegExpCharSet(attributes);
    } else if (name == "interval") {
      regexp = new RegExpInterval(attributes);
    } else if (name == "charrange") {
      regexp = new RegExpCharRange(attributes);
    } else if (name == "intersection") {
      regexp = new RegExpIntersection(parseRegExpContent(e.getChildren(), pwf));
    } else if (name == "concat") {
      regexp = new RegExpConcat(parseRegExpContent(e.getChildren(), pwf));
    } else if (name == "union") {
      regexp = new RegExpUnion(parseRegExpContent(e.getChildren(), pwf));
    } else if (name == "repeat") {
      regexp = new RegExpRepeat(attributes, parseSubRegExp(e, pwf));
    } else if (name == "plus") {
      regexp = new RegExpPlus(parseSubRegExp(e, pwf));
    } else if (name == "optional") {
      regexp = new RegExpOptional(parseSubRegExp(e, pwf));
    } else if (name == "complement") {
      regexp = new RegExpComplement(parseSubRegExp(e, pwf));
    } else if (name == "regexp") {
      if (attributes.getValue("pattern") != null) {
	regexp = new RegExpPattern(attributes);
      } else {
	regexp = new RegExpRef(attributes);
      }
    }
    return regexp;
  }

  private static PwfContent parseRegExpContent(List l, PowerForms pwf) throws ParseException {
    PwfContent content = new PwfContent();

    for (Iterator i = l.iterator(); i.hasNext(); ) {
      RegExp regexp = parseRegExp((Element)i.next(), pwf);
      if (regexp != null) {
	content.add(regexp);
      } else {
	nontermExpected("regexp");
      }
    }
    return content;
  }

  /*
  private static HtmlStatus parseStatus(Element e, PowerForms pwf) throws ParseException {
    HtmlStatus status = null;
    String name = e.getName().intern();
    if (name == "status") {
      status = new HtmlStatus(parseAttributes(e.getAttributes(), pwf));
    }
    return status;
  }
  */

  /* Parsing from Stream */
  /*
    static HtmlDocument parseHtmlDocument(Lexer lexer, PowerForms pwf) throws ParseException {
    HtmlDocument document = new HtmlDocument(pwf);
    HtmlContent content = parseHtmlContent(lexer, pwf);
    document.setContent(content);
    if (lexer.hasMoreTokens()) {
    throw new ParseException(lexer, "unexpected text `"+lexer.getText()+"'");
    }
    return document;
    }

    private static Attributes parseAttributes(Lexer lexer, PowerForms pwf) throws ParseException {
    return parseAttributes(lexer, true, pwf);
    }

    private static Attributes parseAttributes(Lexer lexer, boolean xml, PowerForms pwf) throws ParseException {
    Attributes attributes = new Attributes(pwf);
    while (lexer.token == Token.Name) {
    String name = lexer.getText();
    lexer.match(Token.Name);
    String value = null;
    if (xml || (lexer.token == '=')) {
    lexer.match('=');
    value = lexer.getText();
    lexer.match(Token.AttValue);
    }
    attributes.add(name, value);
    }
    return attributes;
    }

    private static PwfNode parseHtml(Lexer lexer, PowerForms pwf) throws ParseException {
    URL url = lexer.getUrl();
    int line = lexer.getLine();
    PwfNode h = null;
    Attributes attributes = null;
    boolean addID = !true;
    HtmlContent content = new HtmlContent();

    switch (lexer.token) {
    case Token.SOMETHING:
    h = new HtmlSomething(lexer.getText());
    lexer.match(Token.SOMETHING);
    break;
    case Token.FORM_start:
    lexer.match(Token.FORM_start);
    attributes = parseAttributes(lexer, false, pwf);
    if (lexer.token == Token.ENDTAG) {
    lexer.match(Token.ENDTAG);
    } else {
    lexer.match('>');
    content = parseHtmlContent(lexer, pwf);
    lexer.match(Token.FORM_end);
    }
    h = new HtmlForm(attributes, content);
    break;
    case Token.INPUT_start:
    lexer.match(Token.INPUT_start);
    attributes = parseAttributes(lexer, false, pwf);
    if (lexer.token == Token.ENDTAG) {
    lexer.match(Token.ENDTAG);
    } else {
    lexer.match('>');
    if (lexer.token == Token.INPUT_end) {
    lexer.match(Token.INPUT_end);
    }
    }
    h = new HtmlInput(attributes);
    break;
    case Token.SELECT_start:
    lexer.match(Token.SELECT_start);
    attributes = parseAttributes(lexer, false, pwf);
    if (lexer.token == Token.ENDTAG) {
    lexer.match(Token.ENDTAG);
    } else {
    lexer.match('>');
    content = parseHtmlContent(lexer, pwf);
    lexer.match(Token.SELECT_end);
    }
    h = new HtmlSelect(attributes, content);
    break;
    case Token.TEXTAREA_start:
    lexer.match(Token.TEXTAREA_start);
    attributes = parseAttributes(lexer, false, pwf);
    if (lexer.token == Token.ENDTAG) {
    lexer.match(Token.ENDTAG);
    } else {
    lexer.match('>');
    content = parseHtmlContent(lexer, pwf);
    lexer.match(Token.TEXTAREA_end);
    }
    h = new HtmlTextArea(attributes, content);
    break;
    case Token.OPTION_start:
    lexer.match(Token.OPTION_start);
    attributes = parseAttributes(lexer, false, pwf);
    if (lexer.token == Token.ENDTAG) {
    lexer.match(Token.ENDTAG);
    } else {
    lexer.match('>');
    while (lexer.token == Token.SOMETHING) {
    PwfNode h1 = new HtmlSomething(lexer.getText());
    h1.setLocation(lexer.getUrl(), lexer.getLine());
    content.add(h1);
    lexer.match(Token.SOMETHING);
    }
    if (lexer.token == Token.OPTION_end) {
    lexer.match(Token.OPTION_end);
    }
    }
    h = new HtmlOption(attributes, content);
    break;
    case Token.APPLET_start:
    lexer.match(Token.APPLET_start);
    attributes = parseAttributes(lexer, false, pwf);
    if (lexer.token == Token.ENDTAG) {
    lexer.match(Token.ENDTAG);
    } else {
    lexer.match('>');
    content = parseHtmlContent(lexer, pwf);
    lexer.match(Token.APPLET_end);
    }
    h = new HtmlApplet(attributes, content);
    break;
    case Token.RESULT_start:
    lexer.match(Token.RESULT_start);
    attributes = parseAttributes(lexer, pwf);
    if (lexer.token == Token.ENDTAG) {
    lexer.match(Token.ENDTAG);
    } else {
    lexer.match('>');
    if (lexer.token == Token.RESULT_end) {
    lexer.match(Token.RESULT_end);
    }
    }
    h = new HtmlResult(attributes);
    break;
    case Token.SUBMIT_start:
    lexer.match(Token.SUBMIT_start);
    attributes = parseAttributes(lexer, pwf);
    if (lexer.token == Token.ENDTAG) {
    lexer.match(Token.ENDTAG);
    } else {
    lexer.match('>');
    content = parseHtmlContent(lexer, pwf);
    lexer.match(Token.SUBMIT_end);
    }
    h = new HtmlSubmit(attributes, content);
    break;
    default:
    if (firstStatus(lexer.token)) {
    h = parseStatus(lexer, pwf);
    } else {
    throw new ParseException(lexer, "html expected");
    }
    break;
    }
    if ((attributes != null) && addID) {
    attributes.addID();
    }
    h.setLocation(url, line);
    return h;
    }

    private static HtmlContent parseHtmlContent(Lexer lexer, PowerForms pwf) throws ParseException {
    HtmlContent content = new HtmlContent();
    while (firstHtml(lexer.token)) {
    content.add(parseHtml(lexer, pwf));
    }
    return content;
    }

    private static HtmlStatus parseStatus(Lexer lexer, PowerForms pwf) throws ParseException {
    URL url = lexer.getUrl();
    int line = lexer.getLine();
    HtmlStatus status = null;

    switch (lexer.token) {
    case Token.STATUS_start:
    lexer.match(Token.STATUS_start);
    status = new HtmlStatus(parseAttributes(lexer, pwf));
    lexer.matchEndTag(Token.STATUS_end);
    break;
    default:
    throw new ParseException(lexer, "status expected");
    }
    if (status.attributes != null) {
    status.attributes.addID();
    }
    status.setLocation(url, line);
    return status;
    }

    private static boolean firstHtml(int token) {
    switch (token) {
    case Token.SOMETHING:
    case Token.FORM_start:
    case Token.INPUT_start:
    case Token.SELECT_start:
    case Token.TEXTAREA_start:
    case Token.OPTION_start:
    case Token.APPLET_start:
    case Token.RESULT_start:
    case Token.SUBMIT_start:
    return true;
    default:
    return firstStatus(token);
    }
    }

    private static boolean firstStatus(int token) {
    switch (token) {
    case Token.STATUS_start:
    return true;
    default:
    return false;
    }
    }
  */

  /*
    private static Constraint parseConstraint(Lexer lexer) throws ParseException {
    URL url = lexer.getUrl();
    int line = lexer.getLine();
    Constraint constraint = null;
    Attributes attributes = null;

    switch (lexer.token) {
    case Token.CONSTRAINT_start:
    lexer.match(Token.CONSTRAINT_start);
    attributes = parseAttributes(lexer);
    if ((constraint_level > 1) && (attributes.getValue("idref") != null)) {
    constraint = new ConstraintRef(attributes);
    lexer.matchEndTag(Token.CONSTRAINT_end);
    } else {
    lexer.match('>');
    constraint = new ConstraintApply(attributes, parseConstraint(lexer));
    lexer.match(Token.CONSTRAINT_end);
    }
    break;
    case Token.CONSTRAINTDEF_start:
    lexer.match(Token.CONSTRAINTDEF_start);
    attributes = parseAttributes(lexer);
    lexer.match('>');
    constraint = new ConstraintDef(attributes, parseConstraint(lexer));
    lexer.match(Token.CONSTRAINTDEF_end);
    break;
    case Token.IF_start:
    lexer.match(Token.IF_start);
    Expression test = parseExpression(lexer);
    lexer.match(Token.THEN_start);
    Constraint thenpart = parseConstraint(lexer);
    lexer.match(Token.THEN_end);
    lexer.match(Token.ELSE_start);
    Constraint elsepart = parseConstraint(lexer);
    lexer.match(Token.ELSE_end);
    lexer.match(Token.IF_end);
    constraint = new ConstraintIf(test, thenpart, elsepart);
    break;
    case Token.IGNORE_start:
    lexer.match(Token.IGNORE_start);
    attributes = parseAttributes(lexer);
    lexer.matchEndTag(Token.IGNORE_end);
    constraint = new ConstraintIgnore(attributes);
    break;
    default:
    if (firstExpression(lexer.token)) {
    constraint = new ConstraintExpression(parseExpression(lexer));
    } else if (firstRegExp(lexer.token)) {
    Expression exp = new ExpressionMatch(new Attributes(), parseRegExp(lexer));
    exp.setLocation(url, line);
    constraint = new ConstraintExpression(exp);
    } else {
    throw new ParseException(lexer, "constraint expected");
    }
    break;
    }
    if (attributes != null) {
    attributes.addID();
    }
    constraint.setLocation(url, line);
    return constraint;
    }

    private static Expression parseExpression(Lexer lexer) throws ParseException {
    URL url = lexer.getUrl();
    int line = lexer.getLine();
    Expression expression = null;
    Attributes attributes = null;

    switch (lexer.token) {
    case Token.COUNT_start:
    lexer.match(Token.COUNT_start);
    attributes = parseAttributes(lexer);
    lexer.matchEndTag(Token.COUNT_end);
    expression = new ExpressionCount(attributes);
    break;
    case Token.MATCH_start:
    lexer.match(Token.MATCH_start);
    attributes = parseAttributes(lexer);
    lexer.match('>');
    expression = new ExpressionMatch(attributes, parseRegExp(lexer));
    lexer.match(Token.MATCH_end);
    break;
    // case Token.MATCHALL_start:
    // lexer.match(Token.MATCHALL_start);
    // attributes = parseAttributes(lexer);
    // lexer.match('>');
    // expression = new ExpressionMatchAll(attributes, RegparseExp(lexer));
    // lexer.match(Token.MATCHALL_end);
    // break;
    case Token.EQUAL_start:
    lexer.match(Token.EQUAL_start);
    attributes = parseAttributes(lexer);
    lexer.matchEndTag(Token.EQUAL_end);
    expression = new ExpressionEqual(attributes);
    break;
    case Token.FUNCTION_start:
    lexer.match(Token.FUNCTION_start);
    expression = new ExpressionFunction(parseAttributes(lexer));
    lexer.matchEndTag(Token.FUNCTION_end);
    break;
    case Token.AND_start:
    lexer.match(Token.AND_start);
    expression = new ExpressionAnd(parseExpressionContent(lexer));
    lexer.match(Token.AND_end);
    break;
    case Token.OR_start:
    lexer.match(Token.OR_start);
    expression = new ExpressionOr(parseExpressionContent(lexer));
    lexer.match(Token.OR_end);
    break;
    case Token.NOT_start:
    lexer.match(Token.NOT_start);
    expression = new ExpressionNot(parseExpressionContent(lexer));
    lexer.match(Token.NOT_end);
    break;
    default:
    throw new ParseException(lexer, "Expression expected");
    }
    expression.setLocation(url, line);
    if (attributes != null) {
    attributes.addID();
    }
    return expression;
    }

    private static PwfContent parseExpressionContent(Lexer lexer) throws ParseException {
    PwfContent content = new PwfContent();
    while (firstExpression(lexer.token)) {
    content.add(parseExpression(lexer));
    }
    return content;
    }
  */

  /*
    private static Init parseInit(Lexer lexer) throws ParseException {
    URL url = lexer.getUrl();
    int line = lexer.getLine();
    Init init = null;
    Attributes attributes;

    switch (lexer.token) {
    case Token.INIT_start:
    lexer.match(Token.INIT_start);
    attributes = parseAttributes(lexer);
    lexer.matchEndTag(Token.INIT_end);
    init = new Init(attributes);
    break;
    default:
    throw new ParseException(lexer, "init expected");
    }
    init.setLocation(url, line);
    return init;
    }

    private static RegExp parseRegExp(Lexer lexer) throws ParseException {
    URL url = lexer.getUrl();
    int line = lexer.getLine();
    RegExp regexp = null;
    Attributes attributes = null;

    switch (lexer.token) {
    case Token.CONST_start:
    lexer.match(Token.CONST_start);
    attributes = parseAttributes(lexer);
    if (lexer.token == Token.ENDTAG) {
    lexer.match(Token.ENDTAG);
    } else {
    lexer.match('>');
    lexer.match(Token.CONST_end);
    }
    regexp = new RegExpConst(attributes);
    break;
    case Token.EMPTY_start:
    lexer.match(Token.EMPTY_start);
    if (lexer.token == Token.ENDTAG) {
    lexer.match(Token.ENDTAG);
    } else {
    lexer.match('>');
    lexer.match(Token.EMPTY_end);
    }
    regexp = new RegExpEmpty();
    break;
    case Token.ANYCHAR_start:
    lexer.match(Token.ANYCHAR_start);
    if (lexer.token == Token.ENDTAG) {
    lexer.match(Token.ENDTAG);
    } else {
    lexer.match('>');
    lexer.match(Token.ANYCHAR_end);
    }
    regexp = new RegExpAnyChar();
    break;
    case Token.ANYTHING_start:
    lexer.match(Token.ANYTHING_start);
    if (lexer.token == Token.ENDTAG) {
    lexer.match(Token.ENDTAG);
    } else {
    lexer.match('>');
    lexer.match(Token.ANYTHING_end);
    }
    regexp = new RegExpAnything();
    break;
    case Token.CHARSET_start:
    lexer.match(Token.CHARSET_start);
    attributes = parseAttributes(lexer);
    if (lexer.token == Token.ENDTAG) {
    lexer.match(Token.ENDTAG);
    } else {
    lexer.match('>');
    lexer.match(Token.CHARSET_end);
    }
    regexp = new RegExpCharSet(attributes);
    break;
    case Token.INTERVAL_start:
    lexer.match(Token.INTERVAL_start);
    attributes = parseAttributes(lexer);
    if (lexer.token == Token.ENDTAG) {
    lexer.match(Token.ENDTAG);
    } else {
    lexer.match('>');
    lexer.match(Token.INTERVAL_start);
    }
    regexp = new RegExpInterval(attributes);
    break;
    case Token.CHARRANGE_start:
    lexer.match(Token.CHARRANGE_start);
    attributes = parseAttributes(lexer);
    if (lexer.token == Token.ENDTAG) {
    lexer.match(Token.ENDTAG);
    } else {
    lexer.match('>');
    lexer.match(Token.CHARRANGE_end);
    }
    regexp = new RegExpCharRange(attributes);
    break;
    case Token.INTERSECTION_start:
    lexer.match(Token.INTERSECTION_start);
    regexp = new RegExpIntersection(parseRegExpContent(lexer));
    lexer.match(Token.INTERSECTION_end);
    break;
    case Token.CONCAT_start:
    lexer.match(Token.CONCAT_start);
    regexp = new RegExpConcat(parseRegExpContent(lexer));
    lexer.match(Token.CONCAT_end);
    break;
    case Token.UNION_start:
    lexer.match(Token.UNION_start);
    regexp = new RegExpUnion(parseRegExpContent(lexer));
    lexer.match(Token.UNION_end);
    break;
    case Token.REPEAT_start:
    lexer.match(Token.REPEAT_start);
    attributes = parseAttributes(lexer);
    lexer.match('>');
    regexp = new RegExpRepeat(attributes, parseRegExp(lexer));
    lexer.match(Token.REPEAT_end);
    break;
    case Token.PLUS_start:
    lexer.match(Token.PLUS_start);
    regexp = new RegExpPlus(parseRegExp(lexer));
    lexer.match(Token.PLUS_end);
    break;
    case Token.OPTIONAL_start:
    lexer.match(Token.OPTIONAL_start);
    regexp = new RegExpOptional(parseRegExp(lexer));
    lexer.match(Token.OPTIONAL_end);
    break;
    case Token.COMPLEMENT_start:
    lexer.match(Token.COMPLEMENT_start);
    regexp = new RegExpComplement(parseRegExp(lexer));
    lexer.match(Token.COMPLEMENT_end);
    break;
    case Token.REGEXP_start:
    lexer.match(Token.REGEXP_start);
    attributes = parseAttributes(lexer);
    if (attributes.getValue("pattern") != null) {
    regexp = new RegExpPattern(attributes);
    lexer.matchEndTag(Token.REGEXP_end);
    } else {
    regexp = new RegExpRef(attributes);
    lexer.matchEndTag(Token.REGEXP_end);
    }
    break;
    case Token.REGEXPDEF_start:
    lexer.match(Token.REGEXPDEF_start);
    attributes = parseAttributes(lexer);
    lexer.match('>');
    regexp = new RegExpDef(attributes, parseRegExp(lexer));
    lexer.match(Token.REGEXPDEF_end);
    break;
    default:
    throw new ParseException(lexer, "regular expression expected"+": "+Token.getImage(lexer.token));
    }
    if (attributes != null) {
    attributes.addID();
    }
    regexp.setLocation(url, line);
    return regexp;
    }

    private static PwfContent parseRegExpContent(Lexer lexer) throws ParseException {
    PwfContent content = new PwfContent();
    while (firstRegExp(lexer.token)) {
    content.add(parseRegExp(lexer));
    }
    return content;
    }

    static void parsePowerForms(Lexer lexer, PowerForms powerforms) throws ParseException {
    lexer.match(Token.POWERFORMS_start);
    boolean stop = false;
    while (!stop) {
    switch (lexer.token) {
    case Token.REGEXPDEF_start:
    powerforms.addRegExpDef((RegExpDef)parseRegExp(lexer));
    break;
    case Token.CONSTRAINTDEF_start:
    powerforms.addConstraintDef((ConstraintDef)parseConstraint(lexer));
    break;
    case Token.CONSTRAINT_start:
    powerforms.addConstraint((ConstraintApply)parseConstraint(lexer));
    break;
    case Token.INIT_start:
    powerforms.addInitialization(parseInit(lexer));
    break;
    default:
    stop = true;
    break;
    }
    }
    lexer.match(Token.POWERFORMS_end);
    }
  */

  /*
    private static boolean firstConstraint(int token) {
    switch (token) {
    case Token.CONSTRAINT_start:
    case Token.CONSTRAINTDEF_start:
    return true;
    default:
    return false;
    }
    }

    private static boolean firstExpression(int token) {
    switch (token) {
    case Token.COUNT_start:
    case Token.MATCH_start:
    case Token.EQUAL_start:
    case Token.FUNCTION_start:
    case Token.AND_start:
    case Token.OR_start:
    case Token.NOT_start:
    return true;
    default:
    return false;
    }
    }

    private static boolean firstInit(int token) {
    switch (token) {
    case Token.INIT_start:
    return true;
    default:
    return false;
    }
    }

    private static boolean firstRegExp(int token) {
    switch (token) {
    case Token.CONST_start:
    case Token.EMPTY_start:
    case Token.ANYCHAR_start:
    case Token.ANYTHING_start:
    case Token.CHARSET_start:
    case Token.INTERVAL_start:
    case Token.CHARRANGE_start:
    case Token.INTERSECTION_start:
    case Token.CONCAT_start:
    case Token.UNION_start:
    case Token.REPEAT_start:
    case Token.PLUS_start:
    case Token.OPTIONAL_start:
    case Token.COMPLEMENT_start:
    case Token.REGEXP_start:
    case Token.REGEXPDEF_start:
    return true;
    default:
    return false;
    }
    }
  */
}
