/*
 * PowerForms
 * Copyright (C) 1999-2002 BRICS.
 *
 * Author: Mikkel Ricky <ricky@brics.dk>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the  Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 * USA.
 *
 * Home page: http://www.brics.dk/bigwig/powerforms/
 */

package dk.brics.powerforms;

import dk.brics.automaton.*;
import java.io.*;
import java.net.*;
import java.util.*;
import java.text.*;
import org.jdom.*;
import org.jdom.input.*;

public class PowerForms {
  private static final String name = Config.NAME;
  private static final String version = Config.VERSION;

  public static String getVersion() {
    return name+"; version "+version;
  }

  public static final int CRAP_MODE_OPTION = 1024;
  public static final int NO_LOGO_OPTION = 1025;
  public static final int NO_OUTPUT_OPTION = 1026;
  public static final int NO_SCRIPT_OPTION = 1027;
  public static final int MISCURL_OPTION = 1032;
  public static final int STATUS_NONE = 1034;
  public static final int STATUS_IMAGE = 1035;
  public static final int STATUS_CSS = 1036;
  public static final int XHTML_OPTION = 1036;
  public static final int ERRORTITLE_OPTION = 1037;

  private Map regexpdefs, constraintdefs;
  private StringPairMap constraintapplys;
  private StringPairMap dependencies;
  private StringPairMap initializations;
  private ListSet constraints;
  private Set regexps, autocompletes;
  private Map forms, fields;
  private boolean crapMode = false;
  private boolean noLogo = false;
  private boolean noOutput = false;
  private boolean noScript = false;
  private boolean xhtml = true;
  private String miscUrlSpec = null;
  private URL miscUrl = null;
  private String errorTitle = null;
  private Attributes defaultStatusAttributes;
  private ResourceBundle messages;


  public PowerForms() {
    messages = ResourceBundle.getBundle("dk.brics.powerforms.Messages");
    reset();
  }

  public void initialize(Reader in) throws ParseException, TransformException {
    try {
      SAXBuilder builder = new SAXBuilder();
      Document document = builder.build(in);
      initialize(document);
    } catch (JDOMException e) {
      throw new ParseException(e.getMessage());
    }
  }

  private void initialize(Document document) throws ParseException, TransformException {
    Parser.parsePowerForms(document, this);
    check();
  }

  private int pwf_id = 0;

  String getID() {
    return "pwf-id-"+(pwf_id++);
  }

  String getMessage(String key) {
    return messages.getString(key);
  }

  void addRegExpDef(RegExpDef r) throws ParseException {
    if (regexpdefs.containsKey(r.id)) {
      try {
	r.reportError("regexp `"+r.id+"' already defined");
      } catch (TransformException e) {
	throw new ParseException(e.getMessage());
      }
    }
    regexpdefs.put(r.id, r);
    r.setPowerForms(this);
  }

  Set getRegExps() {
    return regexps;
  }

  RegExpDef getRegExpDef(String id) {
    return (RegExpDef)regexpdefs.get(id);
  }

  void addConstraintDef(ConstraintDef c) {
    constraintdefs.put(c.id, c);
    c.setPowerForms(this);
  }

  ConstraintApply getConstraint(String id) {
    for (Iterator i = constraints.iterator(); i.hasNext(); ) {
      ConstraintApply c = (ConstraintApply)i.next();
      if (id.equals(c.id)) {
	return c;
      }
    }
    return null;
  }

  void addRegExp(RegExp regexp) {
    regexps.add(regexp);
  }

  void addConstraint(ConstraintApply c) {
    constraints.add(c);
    c.setPowerForms(this);
  }

  void addInitialization(Init init) {
    initializations.put(init.attributes.getValue("form"), init.attributes.getValue("field"), init.attributes.getValue("value"));
  }

  String getInitialization(String form, String field) {
    return (String)initializations.get(form, field);
  }

  String getInitialization(Object obj) {
    return (String)initializations.get(obj);
  }

  void addConstraintApply(ConstraintApply c) {
    ListSet set = (ListSet)constraintapplys.get(c.formName, c.fieldName);
    if (set == null) {
      set = new ListSet();
      constraintapplys.put(c.formName, c.fieldName, set);
    }
    set.add(c);
  }

  void addAutoComplete(AutoComplete a) {
    autocompletes.add(a);
    a.setPowerForms(this);
  }

  AutoComplete getAutoComplete(String form, String field) {
    ListSet constraints = (ListSet)constraintapplys.get(form, field);
    if (constraints != null) {
      for (Iterator i = constraints.iterator(); i.hasNext(); ) {
	ConstraintApply c = (ConstraintApply)i.next();
	return c.getAutoComplete();
      }
    }
    return null;
  }

  /**
      @return The set of constraints that apply to the field
  */
  ListSet getConstraints(String form, String field) {
    return (ListSet)constraintapplys.get(form, field);
  }

  ListSet getConstraints(Object obj) {
    return (ListSet)constraintapplys.get(obj);
  }

  void addDependency(String form, String field, ConstraintApply c) {
    ListSet set = getDependencies(form, field);
    if (set == null) {
      set = new ListSet();
      dependencies.put(form, field, set);
    }
    set.add(c);
  }

  ListSet getDependencies(String form, String field) {
    return (ListSet)dependencies.get(form, field);
  }

  /**
      @return The set of constraints that depend on the field
  */
  ListSet getDependencies(Object obj) {
    return (ListSet)dependencies.get(obj);
  }

  static String getFormValidateName() {
    return "pwf_form";
  }

  void check() throws TransformException {
    for (Iterator i = regexpdefs.keySet().iterator(); i.hasNext(); ) {
      RegExpDef r = (RegExpDef)regexpdefs.get(i.next());
      r.check();
    }
    for (Iterator i = constraintdefs.keySet().iterator(); i.hasNext(); ) {
      Constraint c = (Constraint)constraintdefs.get(i.next());
      c.check();
    }
    for (Iterator i = constraints.iterator(); i.hasNext(); ) {
      Constraint c = (Constraint)i.next();
      c.check();
    }
    /* Transitive closure like computation on dependency graph */
    boolean cont;
    do {
      cont = false;
      for (Iterator i = dependencies.keySet().iterator(); i.hasNext(); ) {
	Object o = i.next();
	ListSet s = new ListSet();
	s.addAll(getDependencies(o));
	for (Iterator j = getDependencies(o).iterator(); j.hasNext(); ) {
	  ConstraintApply c = (ConstraintApply)j.next();
	  StringPair p = c.getConstrains();
	  ListSet s0 = getDependencies(p);
	  if ((s0 != null) && s.addAll(s0)) {
	    cont = true;
	  }
	}
	dependencies.put(o, s);
      }
    } while (cont);
    for (Iterator i = autocompletes.iterator(); i.hasNext(); ) {
      AutoComplete a = (AutoComplete)i.next();
      a.check();
    }
  }

  String validationError(String key, String argument) {
    String message = getMessage(key);
    return MessageFormat.format(message, new Object[] { argument });
  }

  String validationError(String key, String argument0, String argument1) {
    String message = getMessage(key);
    return MessageFormat.format(message, new Object[] { argument0, argument1 });
  }

  public String toString() {
    CharArrayWriter chars = new CharArrayWriter();
    PrintWriter out = new PrintWriter(chars);
    out.println("<powerforms>");
    out.println("<!-- regexp definitions -->");
    out.println(regexpdefs);
    for (Iterator i = regexpdefs.keySet().iterator(); i.hasNext(); ) {
      Object key = i.next();
      out.println(key+": "+regexpdefs.get(key));
    }
    out.println();

    out.println("<!-- regexps -->");
    for (Iterator i = regexps.iterator(); i.hasNext(); ) {
      out.println(i.next());
    }
    out.println();

    out.println("<!-- constraint definitions -->");
    out.println(constraintdefs);
    for (Iterator i = constraintdefs.keySet().iterator(); i.hasNext(); ) {
      Object key = i.next();
      out.println(key+": "+constraintdefs.get(key));
    }
    out.println();

    out.println("<!-- constraints -->");
    for (Iterator i = constraints.iterator(); i.hasNext(); ) {
      out.println(i.next());
    }
    out.println();

    out.println("<!-- constraint applications -->");
    for (Iterator i = constraintapplys.keySet().iterator(); i.hasNext(); ) {
       Object o = i.next();
       out.println("  "+o+" constrained by");
       for (Iterator j = ((ListSet)constraintapplys.get(o)).iterator(); j.hasNext(); ) {
	 out.println("    "+j.next());
       }
    }
    // out.println(constraintapplys);
    out.println();

    out.println("<!-- dependencies -->");
    for (Iterator i = dependencies.keySet().iterator(); i.hasNext(); ) {
       Object o = i.next();
       out.println("  "+o+" affects ("+getDependencies(o)+")");
       for (Iterator j = getDependencies(o).iterator(); j.hasNext(); ) {
	 out.println("    "+j.next());
       }
    }
    // out.println(dependencies);
    out.println();

    out.println("<!-- initialization -->");
    out.println(initializations);
//    for (Iterator i0 = forms.keySet().iterator(); i0.hasNext(); ) {
//      String form = (String)i0.next();
//      out.print("form: "+form+"; fields: ");
//      Map fieldNames = (Map)forms.get(form);
//      if (fieldNames == null) {
//	out.print(fieldNames);
//      } else {
//	for (Iterator i1 = fieldNames.keySet().iterator(); i1.hasNext(); ) {
//	  out.print(i1.next());
//	  if (i1.hasNext()) {
//	    out.print(", ");
//	  }
//	}
//      }
//      out.println();
//    }
    out.println("</powerforms>");
    out.close();
    return chars.toString();
  }

  String toDot() {
    CharArrayWriter chars = new CharArrayWriter();
    PrintWriter out = new PrintWriter(chars);
    out.println("digraph PowerForms {");
    List fields = new ArrayList();

    for (Iterator i = constraints.iterator(); i.hasNext(); ) {
      ConstraintApply c = (ConstraintApply)i.next();
      if ((c.formName != null) || (c.fieldName != null)) {
	fields.add(new StringPair(c.formName, c.fieldName));
      }
    }
    for (Iterator i = dependencies.keySet().iterator(); i.hasNext(); ) {
      Object key = i.next();
      int index = fields.indexOf(key);
      if (index < 0) {
	fields.add(key);
      }
    }

    for (Iterator i = constraints.iterator(); i.hasNext(); ) {
      ConstraintApply c = (ConstraintApply)i.next();
      out.println("  c"+c.index+" [shape=box, label=\"("+c.formName+", "+c.fieldName+")\"]");

      for (Iterator j = fields.iterator(); j.hasNext(); ) {
	StringPair p = (StringPair)j.next();
	if (true ||
	    ((c.formName == null) || (c.formName.equals(p.first)))
	    && ((c.fieldName == null) || (c.fieldName.equals(p.second)))) {
	  String name = "f"+fields.indexOf(p);
	  ListSet s = getDependencies(p);
	  if ((s != null) && s.contains(c)) {
	    out.println("  c"+c.index+" -> "+name);
	  }
	}
      }
    }

    for (Iterator i = fields.iterator(); i.hasNext(); ) {
      StringPair p = (StringPair)i.next();
      String name = "f"+fields.indexOf(p);
      out.println("  "+name+" [shape=ellipse, label=\"("+p.first+", "+p.second+")\"]");

      ListSet s = getConstraints(p);
      if (s != null) {
	for (Iterator j = s.iterator(); j.hasNext(); ) {
	  ConstraintApply c = (ConstraintApply)j.next();
	  out.println("  "+name+" -> c"+c.index);
	}
      }
      s = getDependencies(p);
      if (s != null) {
	for (Iterator j = s.iterator(); j.hasNext(); ) {
	  ConstraintApply c = (ConstraintApply)j.next();
	  out.println("  "+name+" -> c"+c.index);
	}
      }
    }
    out.println("}");
    out.close();
    return chars.toString();
  }

  public boolean setOption(int option, boolean value) {
    switch (option) {
    case CRAP_MODE_OPTION:
      crapMode = value;
      break;
    case NO_LOGO_OPTION:
      noLogo = true;
      break;
    case NO_OUTPUT_OPTION:
      noOutput = true;
      break;
    case NO_SCRIPT_OPTION:
      noScript = true;
      break;
    case XHTML_OPTION:
      xhtml = true;
      break;
    default:
      return false;
    }
    return true;
  }

  public boolean setOption(int option, String value) throws TransformException {
    switch (option) {
    case MISCURL_OPTION:
      try {
	miscUrlSpec = value;
	miscUrl = new URL(miscUrlSpec);
      } catch (MalformedURLException e) {
	throw new TransformException(e);
      }
      break;
    case ERRORTITLE_OPTION:
      errorTitle = value;
      break;
    default:
      return false;
    }
    return true;
  }

  public Automaton getAutomaton(String id) {
    RegExp r = getRegExpDef(id);
    return r == null ? null : r.getAutomaton();
  }

  String getErrorTitle() {
    return errorTitle;
  }

  void setDefaultStatusAttributes(Attributes attributes) throws ParseException {
    if (defaultStatusAttributes == null) {
      defaultStatusAttributes = new Attributes(this);
    }
    if (attributes != null) {
      String type = attributes.getValue("type");
      if ((type != null) && !type.equals(defaultStatusAttributes.getValue("type"))) {
	// New type; reset all attributes ...
	defaultStatusAttributes.add("type", type);
	if (type.equals("image")) {
	  try {
	    defaultStatusAttributes.add("red", new URL(miscUrl, "red.gif").toString());
	    defaultStatusAttributes.add("yellow", new URL(miscUrl, "yellow.gif").toString());
	    defaultStatusAttributes.add("green", new URL(miscUrl, "green.gif").toString());
	    defaultStatusAttributes.add("na", new URL(miscUrl, "na.gif").toString());
	    defaultStatusAttributes.add("blank", new URL(miscUrl, "blank.gif").toString());
	  } catch (MalformedURLException e) {}
	} else if (type.equals("css")) {
	  defaultStatusAttributes.add("red", "pwfRed");
	  defaultStatusAttributes.add("yellow", "pwfYellow");
	  defaultStatusAttributes.add("green", "pwfGreen");
	  defaultStatusAttributes.add("na", "pwfNa");
	} else if (type.equals("none")) {
	} else {
	  throw new ParseException("unknown default status type: "+type);
	}
      }
      String[] names = { "type", "red", "yellow", "green", "na", "blank" };
      for (int i = 0; i < names.length; i++) {
	String value = attributes.getValue(names[i]);
	if (value != null) {
	  defaultStatusAttributes.add(names[i], value);
	}
      }
    }
  }

  Attributes getDefaultStatusAttributes() {
    if (defaultStatusAttributes == null) {
      try {
	Attributes attributes = new Attributes(this);
	attributes.add("type", "image");
	setDefaultStatusAttributes(attributes);
      } catch (Exception e) {}
    }
    if (defaultStatusAttributes.getValue("type").equals("none")) {
      return null;
    }
    Attributes attributes = new Attributes(this);
    String[] names = { "type", "red", "yellow", "green", "na", "blank" };
    for (int i = 0; i < names.length; i++) {
      attributes.add(names[i], defaultStatusAttributes.getValue(names[i]));
    }
    return attributes;
  }

  URL getMiscUrl() {
    return miscUrl;
  }

  public void reset() {
    regexpdefs = new HashMap();
    constraintdefs = new HashMap();
    constraintapplys = new StringPairMap();
    dependencies = new StringPairMap();
    initializations = new StringPairMap();
    constraints = new ListSet();
    regexps = new HashSet();
    autocompletes = new HashSet();
    forms = new HashMap();
    fields = new HashMap();
    try {
      miscUrlSpec = Config.MISCURL;
      miscUrl = new URL(miscUrlSpec);
    } catch (MalformedURLException e) {}
    /*
    try {
      Attributes attributes = new Attributes(this);
      attributes.add("type", "image");
      setDefaultStatusAttributes(attributes);
    } catch (Exception e) {}
    */
    crapMode = false;
    noLogo = false;
    noOutput = false;
    noScript = false;
    xhtml = false;
    errorTitle = null;
  }

  /*
  public HtmlDocument transform(Reader in, PrintWriter out) throws ParseException, TransformException {
    return transformXHtml(in, out);
  }
  */

  private void checkMiscUrl() throws TransformException {
    if (getMiscUrl() == null) {
      throw new TransformException("malformed miscurl: "+miscUrlSpec);
    }
  }

  public HtmlDocument transform(Reader in, PrintWriter out) throws ParseException, TransformException {
    checkMiscUrl();
    try {
      SAXBuilder builder = new SAXBuilder();
      /* builder.setValidation(true); */
      builder.setIgnoringElementContentWhitespace(false);
      Document document = builder.build(in);
      return transform(document, out);
    } catch (JDOMException e) {
      throw new ParseException(e.getMessage());
    }
  }

  /*
  public HtmlDocument transformHtml(Reader in, PrintWriter out) throws ParseException, TransformException {
    if (xhtml) {
      return transformXHtml(in, out);
    } else {
      return transform(new Lexer(in), out);
    }
  }

  private HtmlDocument transform(Lexer lexer, PrintWriter out) throws ParseException, TransformException {
    checkMiscUrl();
    lexer.setCrapMode(crapMode);
    lexer.nextToken();
    return transform(Parser.parseHtmlDocument(lexer, this), out);
  }
  */

  public HtmlDocument transform(Document document, PrintWriter out) throws ParseException, TransformException {
    return transform(Parser.parseHtmlDocument(document, this), out);
  }

  private boolean errors = false;

  private HtmlDocument transform(HtmlDocument document, PrintWriter out) throws TransformException {
    document.setPowerForms(this);
    document.check();
    if (errors) {
      document.print(out);
      out.flush();
      return null;
    }
    document.code(out);
    out.flush();
    return document;
  }

  static String toScriptString(String s) {
    if (s == null) {
      return "null";
    }
    if (s.indexOf('"') > -1) {
      StringBuffer sb = new StringBuffer();
      for (int i = 0; i < s.length(); i++) {
	if (s.charAt(i) == '"') {
	  sb.append("\\\"");
	} else {
	  sb.append(s.charAt(i));
	}
      }
      return "\""+sb+"\"";
    }
    return "\""+s+"\"";
  }

  void DEBUG(Object o) {
    /* System.err.println(String.valueOf(o)); */
  }

  /*
  static private java.util.logging.Handler handler = null;

  static void log(Object o) {
    if (handler == null) {
      try {
	handler = new java.util.logging.FileHandler("%t/powerforms.log", true);
      } catch (IOException e) {
      } catch (SecurityException e) {
      }
    }
    if (handler != null) {
      handler.publish(new java.util.logging.LogRecord(java.util.logging.Level.ALL, o.toString()));
    }
  }
  */
}

abstract class PwfNode {
  protected URL url = null;
  protected int line = -1;
  protected PowerForms powerforms = null;
  protected PwfElement parent = null;

  void setLocation(URL url, int line) {
    this.url = url;
    this.line = line;
  }

  String getLocation() {
    if ((url == null) && (line < 0)) {
      return null;
    }
    if (url == null) {
      return String.valueOf(line);
    }
    return url+":"+line;
  }

  PwfElement getParent() {
    return parent;
  }

  void setParent(PwfElement parent) {
    this.parent = parent;
  }

  PwfElement getAncestor(Class clazz) {
    for (PwfElement n = getParent(); n != null; n = n.getParent()) {
      if (n.getClass().equals(clazz)) {
	return n;
      }
    }
    return null;
  }
}

abstract class PwfElement extends PwfNode {
  /** EcmaScript generation */
  /* abstract void scriptDefine(HtmlDocument document, PrintWriter out); */
  /* abstract void script(HtmlDocument document, PrintWriter out); */
  /* abstract void scriptInit(HtmlDocument document, PrintWriter out); */

  /** Html generation */
  /* abstract void code(PrintWriter out); */

  /** PowerForms validation */
  /* abstract void codeValidate(Validator validator); */

  /* abstract void print(PrintWriter out); */

  void setPowerForms(PowerForms pwf) {
    this.powerforms = pwf;
    if (content != null) {
      content.setPowerForms(pwf);
    }
  }

  PowerForms getPowerForms() {
    return powerforms;
  }

  protected Attributes attributes;
  protected PwfContent content;
  String id, tagName;

  protected void init(String tagName) {
    init(tagName, null, null);
  }

  protected void init(String tagName, Attributes attributes) {
    init(tagName, attributes, null);
  }

  protected void init(String tagName, PwfContent content) {
    init(tagName, null, content);
  }

  protected void init(String tagName, Attributes attributes, PwfContent content) {
    this.tagName = tagName;
    setAttributes(attributes);
    setContent(content);
  }

  void setAttributes(Attributes attributes) {
    this.attributes = attributes;
    if (this.attributes != null) {
      this.attributes.setElement(this);
      this.id = this.attributes.getValue("id");
    }
  }

  String getTagName() {
    return tagName;
  }

  void setContent(PwfContent content) {
    this.content = content;
    if (content != null) {
      content.setParent(this);
    }
  }

  void addChild(PwfElement e) {
    if (content == null) {
      content = new PwfContent();
    }
    content.add(e);
    e.setParent(this);
  }

  private List errors = null;

  void reportError(String message) throws TransformException {
    if (errors == null) {
      errors = new ArrayList();
    }
    errors.add(message);
    String location = getLocation();
    if (location == null) {
      throw new TransformException(message);
    } else {
      throw new TransformException(getLocation()+": "+message);
    }
  }

  void check() throws TransformException {
    if (content != null) {
      content.check();
    }
  }

  void code(PrintWriter out) {
    if (content != null) {
      content.code(out);
    }
  }

  void scriptDefine(HtmlDocument document, PrintWriter out) {
    if (content != null) {
      content.scriptDefine(document, out);
    }
  }

  void script(HtmlDocument document, PrintWriter out) {
    if (content != null) {
      content.script(document, out);
    }
  }

  void scriptInit(HtmlDocument document, PrintWriter out) {
    if (content != null) {
      content.scriptInit(document, out);
    }
  }

  void print(PrintWriter out) {
    if (errors != null) {
      out.println("<errors>");
      for (int i = 0; i < errors.size(); i++) {
	out.print("<error>");
	out.print(errors.get(i));
	out.println("</error>");
      }
    }
    out.print("<"+tagName);
    if (attributes != null) {
      attributes.print(out);
    }
    if ((content == null) || (content.size() == 0)) {
      out.print("/>");
    } else {
      out.print(">");
      content.print(out);
      out.print("</"+tagName+">");
    }
    if (errors != null) {
      out.println();
      out.print("</errors>");
    }
  }

  public String toString() {
    if (tagName == null) {
      return super.toString();
    }
    return "<"+tagName+"/>";
  }

  void DEBUG(Object o) {
    /* System.err.println(String.valueOf(o)); */
  }
}

class PwfContent {
  protected ArrayList list;

  protected PwfContent() {
    list = new ArrayList();
  }

  void add(PwfNode n) {
    list.add(n);
  }

  int size() {
    return list.size();
  }

  PwfElement get(int i) {
    return (PwfElement)list.get(i);
  }

  void check() throws TransformException {
    for (int i = 0; i < size(); i++) {
      get(i).check();
    }
  }

  void scriptDefine(HtmlDocument document, PrintWriter out) {
    for (int i = 0; i < size(); i++) {
      get(i).scriptDefine(document, out);
    }
  }

  void script(HtmlDocument document, PrintWriter out) {
    for (int i = 0; i < size(); i++) {
      get(i).script(document, out);
    }
  }

  void scriptInit(HtmlDocument document, PrintWriter out) {
    for (int i = 0; i < size(); i++) {
      get(i).scriptInit(document, out);
    }
  }

  void code(PrintWriter out) {
    for (int i = 0; i < size(); i++) {
      get(i).code(out);
    }
  }

  void print(PrintWriter out) {
    for (int i = 0; i < size(); i++) {
      get(i).print(out);
    }
  }

  void setParent(PwfElement element) {
    for (int i = 0; i < size(); i++) {
      get(i).setParent(element);
    }
  }

  void setPowerForms(PowerForms pwf) {
    for (int i = 0; i < size(); i++) {
      get(i).setPowerForms(pwf);
    }
  }
}

class StringPairMap extends HashMap {
  Object put(String s, String t, Object value) {
    StringPair p = new StringPair(s, t);
    return put(p, value);
  }

  Object get(String s, String t) {
    StringPair p = new StringPair(s, t);
    return get(p);
  }
}

class StringPairSet extends ListSet {
  boolean add(String s, String t) {
    StringPair p = new StringPair(s, t);
    return add(p);
  }

  boolean contains(String s, String t) {
    StringPair p = new StringPair(s, t);
    return contains(p);
  }
}

class StringPair {
  String first, second;

  StringPair(String first, String second) {
    this.first = first;
    this.second = second;
  }

  public int hashCode() {
    return (first == null ? 0 : first.hashCode())+(second == null ? 0 : second.hashCode());
  }

  public boolean equals(Object obj) {
    if (obj instanceof StringPair) {
      StringPair p = (StringPair)obj;
      return (first == null ? p.first == null : first.equals(p.first)) &&
	(second == null ? p.second == null : second.equals(p.second));
    }
    return false;
  }

  public String toString() {
    return "("+first+", "+second+")";
  }
}


class ListSet {
  List list;

  ListSet() {
    list = new ArrayList();
  }

  int size() {
    return list.size();
  }

  Iterator iterator() {
    return list.iterator();
  }

  boolean contains(Object o) {
    return list.contains(o);
  }

  boolean add(Object o) {
    if (list.indexOf(o) > -1) {
      return false;
    }
    return list.add(o);
  }

  /*
  boolean addAll(Collection c) {
    boolean change = false;
    for (Iterator i = c.iterator(); i.hasNext(); ) {
      if (add(i.next())) {
	change = true;
      }
    }
    return change;
  }
  */

  boolean addAll(ListSet l) {
    boolean change = false;
    for (Iterator i = l.iterator(); i.hasNext(); ) {
      if (add(i.next())) {
	change = true;
      }
    }
    return change;
  }

  public int hashCode() {
    return list.hashCode();
  }

  public boolean equals(Object o) {
    if (o instanceof ListSet) {
      return list.equals(((ListSet)o).list);
    }
    return false;
  }

  public String toString() {
    return "ListSet {"+list+"}";
  }
}
