/*
 * PowerForms
 * Copyright (C) 1999-2002 BRICS.
 *
 * Author: Mikkel Ricky <ricky@brics.dk>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the  Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 * USA.
 *
 * Home page: http://www.brics.dk/bigwig/powerforms/
 */

package dk.brics.powerforms;

import dk.brics.automaton.*;
import java.io.*;
import java.net.*;
import java.util.*;

abstract class RegExp extends PwfElement {
  static final int CRASH = -1;
  static final int REJECT = 0;
  static final int ACCEPT = 1;

  private static int number = 0;
  private int index;

  protected RegExp() {
    index = number++;
  }

  String getDfaName() {
    return "dfa_"+index;
  }

  protected Automaton automaton = null;

  protected Automaton getAutomaton() {
    if (automaton == null) {
      makeAutomaton();
    }
    return automaton;
  }

  abstract void makeAutomaton();

  void code(PrintWriter out) {}

  private static String format(char c) {
    if (!true) {
      return String.valueOf((int)c);
    } else {
      if ((c < 32) || (127 < c)) {
	String s = Integer.toHexString(c).toUpperCase();
	switch (s.length()) {
	case 1:
	  return "'\\u000"+s+"'";
	case 2:
	  return "'\\u00"+s+"'";
	case 3:
	  return "'\\u0"+s+"'";
	default:
	  return "'\\u"+s+"'";
	}
      }
      if ((c == '\'') || (c == '\\')) {
	return "'\\"+String.valueOf(c)+"'";
      } else {
	return "'"+String.valueOf(c)+"'";
      }
    }
  }

  void scriptInit(HtmlDocument document, PrintWriter out) {}

  void script(HtmlDocument document, PrintWriter out) {
    out.print(getDfaName());
  }

  private int[][] table = null;

  private static int[][] computeTable(RegExp regexp) {
    RunAutomaton r = new RunAutomaton(regexp.getAutomaton());
    char[] c = r.getCharIntervals();
    int[][] table = new int[r.getSize()][];
    int[] transition = new int[3*c.length+1];
    int state = 0;
    for (int i = 0; i < r.getSize(); i++) {
      int s = i;
      if (state == 0) {
	s = r.getInitialState();
	i--;
      } else if (i == r.getInitialState()) {
	continue;
      }
      int offset = 0;
      transition[offset++] = r.isAccept(s) ? 1 : 0;
      for (int j = 0; j < c.length; j++) {
	int next = r.step(s, c[j]);
	if (next != -1) {
	  if (next == r.getInitialState()) {
	    next = 0;
	  } else if (next < r.getInitialState()) {
	    next++;
	  }
	  char high = (j < c.length-1 ? c[j+1] : Character.MAX_VALUE);
	  high--;
	  if ((offset > 1) && (transition[offset-1] == next) && (transition[offset-2] == c[j]-1)) {
	    transition[offset-2] = high;
	  } else {
	    transition[offset++] = c[j];
	    transition[offset++] = high;
	    transition[offset++] = next;
	  }
	}
      }
      table[state] = new int[offset];
      System.arraycopy(transition, 0, table[state], 0, offset);
      state++;
    }
    return table;
  }

  private void compute() {
    if (table == null) {
      table = computeTable(this);
    }
  }

  private String ecmaScriptString = null;

  private String getEcmaScript() {
    if (ecmaScriptString == null) {
      compute();
      StringBuffer buffer = new StringBuffer();
      buffer.append("new Dfa(new Array(");
      for (int i = 0; i < table.length; i++) {
	if (i > 0) {
	  buffer.append(", ");
	}
	buffer.append("new Array("+(table[i][0] == 1));
	for (int j = 1; j < table[i].length; j += 3) {
	  buffer.append(", "+format((char)table[i][j]));
	  buffer.append(", "+format((char)table[i][j+1]));
	  buffer.append(", "+table[i][j+2]);
	}
	buffer.append(")");
      }
      buffer.append("))");
      ecmaScriptString = buffer.toString();
    }
    return ecmaScriptString;
  }

  byte run(String input) {
    compute();
    if ((table == null) || (input == null) || ((table.length == 1) && (table[0][0] == 0))) {
      return CRASH;
    }
    int current_state = 0;
    for (int  i = 0; (i < input.length()) && (current_state >= 0); i++) {
      char c = input.charAt(i);
      int state = current_state;
      current_state = -1;
      for (int j = 1; j < table[state].length; j += 3) {
	if ((table[state][j] <= c) && (c <= table[state][j+1])) {
	  current_state = table[state][j+2];
	  break;
	}
      }
    }
    if (current_state < 0) {
      return CRASH;
    } else if (this.table[current_state][0] == 0) {
      return REJECT;
    } else {
      return ACCEPT;
    }
  }

  protected boolean defined = false;

  void scriptDefine(HtmlDocument document, PrintWriter out) {
    if (!defined) {
      out.println("var "+getDfaName()+" = "+getEcmaScript()+";");
      out.println(getDfaName()+".filter = function(s) { return "+getDfaName()+".run(s) == PowerForms.ACCEPT; }");
      defined = true;
    }
  }
}

class RegExpConst extends RegExp {
  private String value;

  RegExpConst(Attributes attributes) {
    init("const", attributes);
  }

  void check() throws TransformException {
    value = attributes.requireValue("value");
  }

  void makeAutomaton() {
    automaton = Automaton.makeString(value);
  }
}

class RegExpEmpty extends RegExp {
  RegExpEmpty() {
    init("empty");
  }

  String getDfaName() {
    return "Dfa.EMPTY";
  }

  void scriptDefine(HtmlDocument document, PrintWriter out) {}

  void makeAutomaton() {
    automaton = Automaton.makeEmpty();
  }

  byte run(String input) {
    return CRASH;
  }
}

class RegExpAnyChar extends RegExp {
  RegExpAnyChar() {
    init("anychar");
  }

  String getDfaName() {
    return "Dfa.ANYCHAR";
  }

  void scriptDefine(HtmlDocument document, PrintWriter out) {}

  void makeAutomaton() {
    automaton = Automaton.makeAnyChar();
  }
}

class RegExpAnything extends RegExp {
  RegExpAnything() {
    init("anything");
  }

  String getDfaName() {
    return "Dfa.ANYTHING";
  }

  void scriptDefine(HtmlDocument document, PrintWriter out) {}

  void makeAutomaton() {
    automaton = Automaton.makeAnyString();
  }

  byte run(String input) {
    return ACCEPT;
  }
}

class RegExpCharSet extends RegExp {
  private String value;

  RegExpCharSet(Attributes attributes) {
    init("charset", attributes);
  }

  void check() throws TransformException {
    value = attributes.requireValue("value");
  }

  void makeAutomaton() {
    automaton = Automaton.makeCharSet(value);
  }
}

class RegExpCharRange extends RegExp {
  private char low, high;

  RegExpCharRange(Attributes attributes) {
    init("charrange", attributes);
  }

  void check() throws TransformException {
    String value = attributes.requireValue("low");
    low = '\0';
    if (value != null) {
      if (value.length() != 1) {
	reportError("attribute `low' must be a single character");
      } else {
	low = value.charAt(0);
      }
    }
    value = attributes.requireValue("high");
    high = '\0';
    if (value != null) {
      if (value.length() != 1) {
	reportError("attribute `high' must be a single character");
      } else {
	high = value.charAt(0);
      }
    }
    if (low > high) {
      char t = low;
      low = high;
      high = t;
    }
  }

  void makeAutomaton() {
    automaton = Automaton.makeCharRange(low, high);
  }
}

class RegExpInterval extends RegExp {
  private int low, high;
  private int width = 0;
  private int radix = 10;

  RegExpInterval(Attributes attributes) {
    init("interval", attributes);
  }

  void check() throws TransformException {
    width = attributes.getIntValue("width", 0);
    if (width < 0) {
      width = 0;
    }
    radix = attributes.getIntValue("radix", 10);
    if (radix < 0) {
      radix = 10;
    } else if (radix > Character.MAX_RADIX) {
      reportError("value of `radix' must be less than "+Character.MAX_RADIX);
    }
    if (radix <= Character.MAX_RADIX) {
      String s = attributes.requireValue("low");
      if (s != null) {
	try {
	  low = Integer.parseInt(s, radix);
	} catch (NumberFormatException e) {
	  reportError("value of `low' must be an integer in base "+radix);
	}
      }
      s = attributes.requireValue("high");
      if (s != null) {
	try {
	  high = Integer.parseInt(s, radix);
	} catch (NumberFormatException e) {
	  reportError("value of `high' must be an integer in base "+radix);
	}
      }
      if (low > high) {
	int t = low;
	low = high;
	high = t;
      }
    }
  }

  void makeAutomaton() {
    automaton = interval(low, high, radix, width);
    automaton.minimize();
  }

  private static String zeropad(int number, int radix, int width) {
    boolean negative = (number < 0);
    if (negative) {
      number = -number;
    }
    StringBuffer sb = new StringBuffer();
    sb.append(Integer.toString(number, radix));
    while (sb.length() < width) {
      sb.insert(0, '0');
    }
    if (negative) {
      sb.insert(0, '-');
    }
    return sb.toString();
  }

  private static Automaton stringRange(String low, String high) {
    if (low.length() != high.length()) {
      return Automaton.makeEmpty();
    }
    if (low.length() == 0) {
      return Automaton.makeEmptyString();
    }
    Automaton a = Automaton.makeCharRange(low.charAt(0), high.charAt(0));
    for (int i = 1; i < low.length(); i++) {
      a = a.concatenate(Automaton.makeCharRange(low.charAt(i), high.charAt(i)));
    }
    return a;
  }

  static Automaton interval(int low, int high, int radix, int width) {
    if (low == high) {
      return Automaton.makeString(zeropad(low, radix, width));
    }
    if (low > high) {
      int tmp = low;
      low = high;
      high = tmp;
    }
    if ((low < 0) && (low == -high)) {
      return Automaton.makeString("-").optional().concatenate(interval(1, high, width, radix)).union(Automaton.makeString(zeropad(0, radix, width)));
    }
    if (low < 0) {
      if (high < 0) {
	return Automaton.makeString("-").concatenate(interval(-low, -high, radix, width));
      } else {
	return Automaton.makeString("-").concatenate(interval(-low, 1, radix, width)).union(interval(0, high, radix, width));
      }
    } else {
      int digit = low%radix;
      int tmp = low;
      int base = 1;
      for (int k = digit; (k < radix-1) && (tmp+base <= high); k++, tmp += base);
      if (digit == 0) {
	base *= radix;
	int low_length = Integer.toString(low, radix).length();
	while ((--low_length > 0) && (tmp+base < high)) {
	  for (int k = (low/base)%radix; (k < radix-1) && (tmp+base <= high); k++, tmp += base);
	  base *= radix;
	}
      }
      if (low == tmp) {
	if (tmp+1 <= high) {
	  return Automaton.makeString(zeropad(low, radix, width)).union(interval(tmp+1, high, radix, width));
	} else {
	  return Automaton.makeString(zeropad(low, radix, width));
	}
      } else {
	String l = zeropad(low, radix, width);
	String h = zeropad(tmp, radix, width);
	Automaton res = stringRange(l, h);
	if (tmp+1 <= high) {
	  res = res.union(interval(tmp+1, high, radix, width));
	}
	return res;
      }
    }
  }
}

class RegExpRef extends RegExp {
  private String idref;
  private URL url;
  private RegExpDef def;

  RegExpRef(Attributes attributes) {
    init("regexp", attributes);
  }

  void check() throws TransformException {
    String idref = attributes.getValue("idref");
    if (idref != null) {
      def = getPowerForms().getRegExpDef(idref);
      if (def == null) {
	reportError("regexp `"+idref+"' not defined");
      }
    } else {
      String spec = attributes.getValue("url");
      if (spec != null) {
	try {
	  url = new URL(spec);
	  automaton = Automaton.load(url);
	} catch (MalformedURLException e) {
	  reportError("attribute `url' must be a valid filename or url");
	} catch (Exception e) {
	  reportError("unable to import regular expression from `"+url+"'");
	}
      }
    }
    if ((idref == null) && (url == null)) {
      reportError("missing attribute `idref' or 'url' for `"+tagName+"'");
    }
  }

  String getDfaName() {
    return def != null ? def.getDfaName() : super.getDfaName();
  }

  void scriptDefine(HtmlDocument document, PrintWriter out) {
    if (def != null) {
      def.scriptDefine(document, out);
    } else {
      super.scriptDefine(document, out);
    }
  }

  void makeAutomaton() {
    if (def != null) {
      automaton = def.getAutomaton();
    }
  }
}

class RegExpPattern extends RegExp {
  private String pattern;

  RegExpPattern(Attributes attributes) {
    init("regexp", attributes);
  }

  void check() throws TransformException {
    pattern = attributes.requireValue("pattern");
    if (pattern != null) {
      if (pattern.length() == 0) {
	automaton = Automaton.makeEmptyString();
      } else try {
	automaton = new dk.brics.automaton.RegExp(pattern).toAutomaton();
      } catch (IllegalArgumentException e) {
	reportError("syntax error in pattern: "+e.getMessage());
      }
    }
  }

  void makeAutomaton() {}
}

class RegExpDef extends RegExp {
  RegExp regexp;

  RegExpDef(Attributes attributes, RegExp regexp) {
    this.regexp = regexp;
    PwfContent content = new PwfContent();
    content.add(regexp);
    init("regexp", attributes, content);
  }

  void check() throws TransformException {
    attributes.requireValue("id");
    super.check();
  }

  String getDfaName() {
    return regexp.getDfaName();
  }

  void scriptDefine(HtmlDocument document, PrintWriter out) {
    if (!defined) {
      // out.println("/* RegExp definition: "+id+" */");
      regexp.scriptDefine(document, out);
      defined = true;
    }
  }

  void makeAutomaton() {
    automaton = regexp.getAutomaton();
  }

  void code(PrintWriter out) {}
}

class RegExpIntersection extends RegExp {
  RegExpIntersection(PwfContent regexps) {
    init("intersection", regexps);
  }

  void makeAutomaton() {
    if (content.size() == 0) {
      automaton = Automaton.makeEmpty();
    } else {
      automaton = ((RegExp)content.get(0)).getAutomaton();
      for (int i = 1; i < content.size(); i++) {
	automaton = automaton.intersection(((RegExp)content.get(i)).getAutomaton());
      }
      automaton.minimize();
    }
  }
}

class RegExpConcat extends RegExp {
  RegExpConcat(PwfContent regexps) {
    init("concat", regexps);
  }

  void makeAutomaton() {
    if (content.size() == 0) {
      automaton = Automaton.makeEmpty();
    } else {
      automaton = ((RegExp)content.get(0)).getAutomaton();
      for (int i = 1; i < content.size(); i++) {
	automaton = automaton.concatenate(((RegExp)content.get(i)).getAutomaton());
      }
      automaton.minimize();
    }
  }
}

class RegExpUnion extends RegExp {
  RegExpUnion(PwfContent regexps) {
    init("union", regexps);
  }

  void makeAutomaton() {
    if (content.size() == 0) {
      automaton = Automaton.makeEmpty();
    } else {
      automaton = ((RegExp)content.get(0)).getAutomaton();
      for (int i = 1; i < content.size(); i++) {
	automaton = automaton.union(((RegExp)content.get(i)).getAutomaton());
      }
      automaton.minimize();
    }
  }
}

class RegExpPlus extends RegExp {
  private RegExp regexp;

  RegExpPlus(RegExp regexp) {
    this.regexp = regexp;
    PwfContent content = new PwfContent();
    content.add(regexp);
    init("plus", content);
  }

  void makeAutomaton() {
    automaton = regexp.getAutomaton().repeat(1);
    automaton.minimize();
  }
}

class RegExpOptional extends RegExp {
  private RegExp regexp;

  RegExpOptional(RegExp regexp) {
    this.regexp = regexp;
    PwfContent content = new PwfContent();
    content.add(regexp);
    init("optional", content);
  }

  void makeAutomaton() {
    automaton = regexp.getAutomaton().optional();
    automaton.minimize();
  }
}

class RegExpRepeat extends RegExp {
  private RegExp regexp;
  private int min, max;

  RegExpRepeat(Attributes attributes, RegExp regexp) {
    this.regexp = regexp;
    PwfContent content = new PwfContent();
    content.add(regexp);
    init("repeat", attributes, content);
  }

  void check() throws TransformException {
    String value = attributes.getValue("count");
    if (value != null) {
      min = attributes.getIntValue("count");
      if (min < 0) {
	reportError("attribute `count' must be an non-negative integer");
      } else {
	max = min;
      }
    } else {
      min = attributes.getIntValue("min", 0);
      if (min < 0) {
	reportError("attribute `min' must be an non-negative integer");
      }
      max = attributes.getIntValue("max", -1);
      if ((max < 0) && (attributes.getValue("max") != null)) {
	reportError("attribute `max' must be an non-negative integer");
      }
      if ((0 <= max) && (max < min)) {
	int t = min;
	min = max;
	max = t;
      }
    }
    regexp.check();
  }

  void makeAutomaton() {
    if (max == -1) {
      automaton = regexp.getAutomaton().repeat(min);
    } else {
      automaton = regexp.getAutomaton().repeat(min, max);
    }
    automaton.minimize();
  }
}

class RegExpComplement extends RegExp {
  private RegExp regexp;

  RegExpComplement(RegExp regexp) {
    this.regexp = regexp;
    PwfContent content = new PwfContent();
    content.add(regexp);
    init("complement", content);
  }

  void makeAutomaton() {
    automaton = regexp.getAutomaton().complement();
    automaton.minimize();
  }
}
